# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import glob
import hashlib
import logging
import operator
import os
import re
from configparser import RawConfigParser
from html.parser import HTMLParser
from typing import IO, TYPE_CHECKING, Any, Optional

import gfal2
import requests

from rucio.common.config import get_config_dirs
from rucio.common.constants import RseAttr
from rucio.common.dumper import DUMPS_CACHE_DIR, HTTPDownloadFailed, ddmendpoint_url, gfal_download_to_file, http_download_to_file, temp_file
from rucio.core.credential import get_signed_url
from rucio.core.rse import get_rse_id, list_rse_attributes

if TYPE_CHECKING:
    from collections.abc import Iterable

CHUNK_SIZE = 10485760

__DUMPERCONFIGDIRS = list(
    filter(
        os.path.exists,
        (
            os.path.join(confdir, 'auditor') for confdir in get_config_dirs()
        )
    )
)

OBJECTSTORE_NUM_TRIES = 30


class Parser(RawConfigParser):
    '''
    RawConfigParser subclass that doesn't modify the the name of the options
    and removes any quotes around the string values.
    '''
    remove_quotes_re = re.compile(r"^'(.+)'$")
    remove_double_quotes_re = re.compile(r'^"(.+)"$')

    def optionxform(
            self,
            optionstr: str
    ) -> str:
        return optionstr

    def get(
            self,
            section: str,
            option: str
    ) -> Any:
        value = super(Parser, self).get(section, option)
        if isinstance(value, str):
            value = self.remove_quotes_re.sub(r'\1', value)
            value = self.remove_double_quotes_re.sub(r'\1', value)
        return value

    def items(self, section):
        return [(name, self.get(section, name)) for name in self.options(section)]


def mkdir(dir_: str) -> None:
    '''
    This functions creates the `dir` directory if it doesn't exist. If `dir`
    already exists this function does nothing.
    '''
    try:
        os.mkdir(dir_)
    except OSError as error:
        if error.errno != 17:
            raise error


def get_newest(
        base_url: str,
        url_pattern: str,
        links: "Iterable[str]"
) -> tuple[str, datetime.datetime]:
    '''
    Returns a tuple with the newest url in the `links` list matching the
    pattern `url_pattern` and a datetime object representing the creation
    date of the url.

    The creation date is extracted from the url using datetime.strptime().
    '''
    logger = logging.getLogger('auditor.srmdumps')
    times = []

    pattern_components = url_pattern.split('/')

    date_pattern = '{0}/{1}'.format(base_url, pattern_components[0])
    if len(pattern_components) > 1:
        postfix = '/' + '/'.join(pattern_components[1:])
    else:
        postfix = ''

    for link in links:
        try:
            time = datetime.datetime.strptime(link, date_pattern)
        except ValueError:
            pass
        else:
            times.append((str(link) + postfix, time))

    if not times:
        msg = 'No links found matching the pattern {0} in {1}'.format(date_pattern, links)
        logger.error(msg)
        raise RuntimeError(msg)

    return max(times, key=operator.itemgetter(1))


def gfal_links(base_url: str) -> list[str]:
    '''
    Returns a list of the urls contained in `base_url`.
    '''
    ctxt = gfal2.creat_context()  # pylint: disable=no-member
    return ['/'.join((base_url, f)) for f in ctxt.listdir(str(base_url))]


class _LinkCollector(HTMLParser):
    def __init__(self):
        super(_LinkCollector, self).__init__()
        self.links = []

    def handle_starttag(
            self, tag: str,
            attrs: "Iterable[tuple[str, str]]"
    ) -> None:
        if tag == 'a':
            self.links.append(
                next(value for key, value in attrs if key == 'href')
            )


def http_links(base_url: str) -> list[str]:
    '''
    Returns a list of the urls contained in `base_url`.
    '''
    html = requests.get(base_url).text
    link_collector = _LinkCollector()

    link_collector.feed(html)
    links = []
    for link in link_collector.links:
        if not link.startswith('http://') and not link.startswith('https://'):
            links.append('{0}/{1}'.format(base_url, link))
        else:
            links.append(link)
    return links


protocol_funcs = {
    'davs': {
        'links': gfal_links,
        'download': gfal_download_to_file,
    },
    'gsiftp': {
        'links': gfal_links,
        'download': gfal_download_to_file,
    },
    'root': {
        'links': gfal_links,
        'download': gfal_download_to_file,
    },
    'srm': {
        'links': gfal_links,
        'download': gfal_download_to_file,
    },
    'http': {
        'links': http_links,
        'download': http_download_to_file,
    },
    'https': {
        'links': http_links,
        'download': http_download_to_file,
    },
}


def protocol(url: str) -> str:
    '''
    Given the URL `url` returns a string with the protocol part.
    '''
    proto = url.split('://')[0]
    if proto not in protocol_funcs:
        raise RuntimeError('Protocol {0} not supported'.format(proto))

    return proto


def get_links(base_url: str) -> list[str]:
    '''
    Given the URL `base_url` returns the URLs linked or contained in it.
    '''
    return protocol_funcs[protocol(base_url)]['links'](base_url)


def download(url: str, filename: IO) -> None:
    '''
    Given the URL `url` downloads its contents on `filename`.
    '''
    return protocol_funcs[protocol(url)]['download'](url, filename)


def parse_configuration(conf_dirs: Optional[list[str]] = None) -> Parser:
    '''
    Parses the configuration for the endpoints contained in `conf_dir`.
    Returns a ConfParser.RawConfParser subclass instance.
    '''
    conf_dirs = conf_dirs or __DUMPERCONFIGDIRS
    logger = logging.getLogger('auditor.srmdumps')
    if len(conf_dirs) == 0:
        logger.error('No configuration directory given to load SRM dumps paths')
        raise Exception('No configuration directory given to load SRM dumps paths')

    configuration = Parser({
        'disabled': False,
    })

    for conf_dir in conf_dirs:
        configuration.read(glob.glob(conf_dir + '/*.cfg'))
    return configuration


def download_rse_dump(
        rse: str,
        configuration: RawConfigParser,
        date: Optional[datetime.datetime] = None,
        destdir: str = DUMPS_CACHE_DIR
) -> tuple[str, datetime.datetime]:
    '''
    Downloads the dump for the given ddmendpoint. If this endpoint does not
    follow the standardized method to publish the dumps it should have an
    entry in the `configuration` object describing how to download the dump.

    `rse` is the DDMEndpoint name.

    `configuration` is a RawConfigParser subclass.

    `date` is a datetime instance with the date of the desired dump or None
    to download the latest available dump.

    `destdir` is the directory where the dump will be saved (the final component
    in the path is created if it doesn't exist).

    Return value: a tuple with the filename and a datetime instance with
    the date of the dump.
    '''
    logger = logging.getLogger('auditor.srmdumps')
    base_url, url_pattern = generate_url(rse, configuration)

    if not os.path.isdir(destdir):
        os.mkdir(destdir)

    # check for objectstores, which need to be handled differently
    rse_id = get_rse_id(rse)
    rse_attr = list_rse_attributes(rse_id)
    if RseAttr.IS_OBJECT_STORE in rse_attr and rse_attr[RseAttr.IS_OBJECT_STORE] is not False:
        tries = 1
        if date is None:
            # on objectstores, can't list dump files, so try the last N dates
            date = datetime.datetime.now()
            tries = OBJECTSTORE_NUM_TRIES
        path = ''
        while tries > 0:
            url = '{0}/{1}'.format(base_url, date.strftime(url_pattern))

            filename = '{0}_{1}_{2}_{3}'.format(
                'ddmendpoint',
                rse,
                date.strftime('%d-%m-%Y'),
                hashlib.sha1(url.encode()).hexdigest()
            )
            filename = re.sub(r'\W', '-', filename)
            path = os.path.join(destdir, filename)
            if not os.path.exists(path):
                logger.debug('Trying to download: "%s"', url)
                if RseAttr.SIGN_URL in rse_attr:
                    url = get_signed_url(rse_id, rse_attr[RseAttr.SIGN_URL], 'read', url)
                try:
                    with temp_file(destdir, final_name=filename) as (f, _):
                        download(url, f)
                    tries = 0
                except (HTTPDownloadFailed, gfal2.GError):
                    tries -= 1
                    date = date - datetime.timedelta(1)
    else:
        if date is None:
            logger.debug('Looking for site dumps in: "%s"', base_url)
            links = get_links(base_url)
            url, date = get_newest(base_url, url_pattern, links)
        else:
            url = '{0}/{1}'.format(base_url, date.strftime(url_pattern))

        filename = '{0}_{1}_{2}_{3}'.format(
            'ddmendpoint',
            rse,
            date.strftime('%d-%m-%Y'),
            hashlib.sha1(url.encode()).hexdigest()
        )
        filename = re.sub(r'\W', '-', filename)
        path = os.path.join(destdir, filename)

        if not os.path.exists(path):
            logger.debug('Trying to download: "%s"', url)
            with temp_file(destdir, final_name=filename) as (f, _):
                download(url, f)

    return (path, date)


def generate_url(
        rse: str,
        config: RawConfigParser
) -> tuple[str, str]:
    '''
    :param rse: Name of the endpoint.
    :param config: RawConfigParser instance which may have configuration
    related to the endpoint.
    :returns: Tuple with the URL where the links can be queried to find new
    dumps and the pattern used to parse the date of the dump of the files/directories
    listed..
    '''
    site = rse.split('_')[0]
    if site not in config.sections():
        base_url = ddmendpoint_url(rse) + 'dumps'
        url_pattern = 'dump_%Y%m%d'
    else:
        url_components = config.get(site, rse).split('/')
        # The pattern may not be the last component
        pattern_index = next(idx for idx, comp in enumerate(url_components) if '%m' in comp)
        base_url = '/'.join(url_components[:pattern_index])
        url_pattern = '/'.join(url_components[pattern_index:])

    return base_url, url_pattern



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
from typing import TYPE_CHECKING

import pymongo

from rucio.common import config, exception
from rucio.common.types import InternalScope
from rucio.core.did_meta_plugins.did_meta_plugin_interface import DidMetaPlugin
from rucio.core.did_meta_plugins.filter_engine import FilterEngine

if TYPE_CHECKING:
    from typing import Optional

    from sqlalchemy.orm import Session

IMMUTABLE_KEYS = [
    '_id',  # index key
    'scope',  # generated on insert
    'name',  # generated on insert
    'vo'  # generated on insert
]


class MongoDidMeta(DidMetaPlugin):
    def __init__(
        self,
        host: "Optional[str]" = None,
        port: "Optional[int]" = None,
        db: "Optional[str]" = None,
        collection: "Optional[str]" = None,
        user: "Optional[str]" = None,
        password: "Optional[str]" = None,
    ):
        super(MongoDidMeta, self).__init__()

        # Validate required parameters.
        con_params = {
            'mongo_service_host': host,
            'mongo_service_port': port,
            'mongo_db': db,
            'mongo_collection': collection,
        }

        for param in con_params:
            if con_params[param] is None:
                if config.config_has_option('metadata', param):
                    con_params[param] = (
                        config.config_get_int('metadata', param)
                        if param == 'mongo_service_port'
                        else config.config_get('metadata', param)
                    )
                else:
                    raise exception.ConnectionParameterNotFound(param)

        if user is None and config.config_has_option('metadata', 'mongo_user'):
            user = config.config_get('metadata', 'mongo_user', default=None)

        if password is None and config.config_has_option('metadata', 'mongo_password'):
            password = config.config_get('metadata', 'mongo_password')

        # Set the auth (fallback to an anonymous connection if either user or password is not defined).
        auth = "" if not user or not password else f"{user}:{password}@"

        self.client = pymongo.MongoClient(
            f"mongodb://{auth}{con_params['mongo_service_host']}:{con_params['mongo_service_port']}/"
        )
        self.db = self.client[con_params['mongo_db']]
        self.col = self.db[con_params['mongo_collection']]

        self.plugin_name = "MONGO"

    def drop_database(self):
        self.client.drop_database(self.db.name)

    def get_metadata(self, scope, name, *, session: "Optional[Session]" = None):
        """
        Get data identifier metadata.

        :param scope: The scope name
        :param name: The data identifier name
        :param session: The database session in use
        :returns: The metadata for the did
        """
        # get first document with this did == _id
        doc = self.col.find_one({
            "_id": "{}:{}".format(scope.internal, name)
        })

        # pop immutable keys
        for key in IMMUTABLE_KEYS:
            if key in doc:
                doc.pop(key)

        if not doc:
            raise exception.DataIdentifierNotFound(f"No metadata found for did '{scope}:{name}'")
        return doc

    def set_metadata(self, scope, name, key, value, recursive=False, *, session: "Optional[Session]" = None):
        """
        Set single metadata key.

        :param scope: the scope of did
        :param name: the name of the did
        :param key: the key to be added
        :param value: the value of the key to be added
        :param recursive: recurse into DIDs (not supported)
        :param session: The database session in use
        """
        self.set_metadata_bulk(scope=scope, name=name, metadata={key: value}, recursive=recursive, session=session)

    def set_metadata_bulk(self, scope, name, metadata, recursive=False, *, session: "Optional[Session]" = None):
        """
        Bulk set metadata keys.

        :param scope: the scope of did
        :param name: the name of the did
        :param metadata: dictionary of metadata keypairs to be added
        :param recursive: recurse into DIDs (not supported)
        :param session: The database session in use
        """
        # pop immutable keys
        for key in IMMUTABLE_KEYS:
            if key in metadata:
                metadata.pop(key)

        # set first document with did == _id
        self.col.update_one(
            {
                "_id": "{}:{}".format(scope.internal, name)
            },
            {
                '$set': metadata,
                '$setOnInsert': {
                    'scope': "{}".format(scope.external),
                    'vo': "{}".format(scope.vo),
                    'name': "{}".format(name)
                }
            },
            upsert=True
        )

    def delete_metadata(self, scope, name, key, *, session: "Optional[Session]" = None):
        """
        Delete a key from metadata.

        :param scope: the scope of did
        :param name: the name of the did
        :param key: the key to be deleted
        """
        meta = {key: ""}
        try:
            self.col.update_one({"_id": "{}:{}".format(scope.internal, name)}, {'$unset': meta})
        except Exception as e:
            raise exception.DataIdentifierNotFound(e)

    def list_dids(self, scope, filters, did_type='collection', ignore_case=False, limit=None,
                  offset=None, long=False, recursive=False, ignore_dids=None, *, session: "Optional[Session]" = None):
        if not ignore_dids:
            ignore_dids = set()

        # backwards compatibility for filters as single {}.
        if isinstance(filters, dict):
            filters = [filters]

        # instantiate fe and create mongo query
        fe = FilterEngine(filters, model_class=None, strict_coerce=False)
        mongo_query_str = fe.create_mongo_query(
            additional_filters=[
                ('scope', operator.eq, scope.internal),
                ('vo', operator.eq, scope.vo)
            ]
        )

        if recursive:
            # TODO: possible, but requires retrieving the results of a concurrent sqla query to call list_content on for datasets and containers
            raise exception.UnsupportedOperation(
                "'{}' metadata module does not currently support recursive searches".format(
                    self.plugin_name.lower()
                ))

        if long:
            query_result = self.col.find(mongo_query_str)
            if limit:
                query_result = query_result.limit(limit)
            for did in query_result:
                did_full = "{}:{}".format(did['scope'], did['name'])
                if did_full not in ignore_dids:  # aggregating recursive queries may contain duplicate DIDs
                    ignore_dids.add(did_full)
                    yield {
                        'scope': InternalScope(did['scope']),
                        'name': did['name'],
                        'did_type': "N/A",
                        'bytes': "N/A",
                        'length': "N/A"
                    }
        else:
            query_result = self.col.find(mongo_query_str)
            if limit:
                query_result = query_result.limit(limit)
            for did in query_result:
                did_full = "{}:{}".format(did['scope'], did['name'])
                if did_full not in ignore_dids:  # aggregating recursive queries may contain duplicate DIDs
                    ignore_dids.add(did_full)
                    yield did['name']

    def manages_key(self, key, *, session: "Optional[Session]" = None):
        return True

    def get_plugin_name(self):
        """
        Returns a unique identifier for this plugin. This can be later used for filtering down results to this plugin only.

        :returns: The name of the plugin
        """
        return self.plugin_name



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' add clob payload to messages '''

import sqlalchemy as sa
from alembic import context
from alembic.op import add_column, drop_column

# Alembic revision identifiers
revision = '90f47792bb76'
down_revision = 'bf3baa1c1474'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        add_column('messages', sa.Column('payload_nolimit', sa.Text), schema=schema)
        add_column('messages_history', sa.Column('payload_nolimit', sa.Text), schema=schema)


def downgrade():
    '''
    Downgrade the database to the previous revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        drop_column('messages', 'payload_nolimit', schema=schema)
        drop_column('messages_history', 'payload_nolimit', schema=schema)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Judge-Injector is a daemon to asynchronously create replication rules
"""
import functools
import logging
import threading
import time
from copy import deepcopy
from datetime import datetime, timedelta
from random import randint
from re import match
from typing import TYPE_CHECKING, Optional

from sqlalchemy.exc import DatabaseError

import rucio.db.sqla.util
from rucio.common.exception import DatabaseException, InsufficientAccountLimit, ReplicationRuleCreationTemporaryFailed, RSEWriteBlocked, RuleNotFound
from rucio.common.logging import setup_logging
from rucio.core.monitor import MetricManager
from rucio.core.rule import get_injected_rules, inject_rule, update_rule
from rucio.daemons.common import HeartbeatHandler, run_daemon
from rucio.db.sqla.constants import ORACLE_CONNECTION_LOST_CONTACT_REGEX, ORACLE_RESOURCE_BUSY_REGEX

if TYPE_CHECKING:
    from types import FrameType

METRICS = MetricManager(module=__name__)
graceful_stop = threading.Event()
DAEMON_NAME = 'judge-injector'


def rule_injector(
        once: bool = False,
        sleep_time: int = 60
) -> None:
    """
    Main loop to check for asynchronous creation of replication rules
    """
    paused_rules = {}  # {rule_id: datetime}
    run_daemon(
        once=once,
        graceful_stop=graceful_stop,
        executable=DAEMON_NAME,
        partition_wait_time=1,
        sleep_time=sleep_time,
        run_once_fnc=functools.partial(
            run_once,
            paused_rules=paused_rules,
        )
    )


def run_once(
        paused_rules: dict[str, datetime],
        heartbeat_handler: HeartbeatHandler,
        **_kwargs
) -> None:
    worker_number, total_workers, logger = heartbeat_handler.live()

    start = time.time()

    # Refresh paused rules
    iter_paused_rules = deepcopy(paused_rules)
    for key in iter_paused_rules:
        if datetime.utcnow() > paused_rules[key]:
            del paused_rules[key]

    rules = get_injected_rules(total_workers=total_workers,
                               worker_number=worker_number,
                               limit=100,
                               blocked_rules=[key for key in paused_rules])
    logger(logging.DEBUG, 'index query time %f fetch size is %d' % (time.time() - start, len(rules)))

    if not rules:
        logger(logging.DEBUG, 'did not get any work (paused_rules=%s)' % str(len(paused_rules)))
        return

    for rule_id in rules:
        _, _, logger = heartbeat_handler.live()
        logger(logging.INFO, 'Injecting rule %s' % rule_id)
        if graceful_stop.is_set():
            break
        try:
            start = time.time()
            inject_rule(rule_id=rule_id, logger=logger)
            logger(logging.DEBUG, 'injection of %s took %f' % (rule_id, time.time() - start))
        except (DatabaseException, DatabaseError) as e:
            if match(ORACLE_RESOURCE_BUSY_REGEX, str(e.args[0])):
                paused_rules[rule_id] = datetime.utcnow() + timedelta(seconds=randint(60, 600))  # noqa: S311
                METRICS.counter('exceptions.{exception}').labels(exception='LocksDetected').inc()
                logger(logging.WARNING, 'Locks detected for %s' % rule_id)
            elif match('.*QueuePool.*', str(e.args[0])):
                logger(logging.WARNING, 'DatabaseException', exc_info=True)
                METRICS.counter('exceptions.{exception}').labels(exception=e.__class__.__name__).inc()
            elif match(ORACLE_CONNECTION_LOST_CONTACT_REGEX, str(e.args[0])):
                logger(logging.WARNING, 'DatabaseException', exc_info=True)
                METRICS.counter('exceptions.{exception}').labels(exception=e.__class__.__name__).inc()
            else:
                logger(logging.ERROR, 'DatabaseException', exc_info=True)
                METRICS.counter('exceptions.{exception}').labels(exception=e.__class__.__name__).inc()
        except (RSEWriteBlocked) as e:
            paused_rules[rule_id] = datetime.utcnow() + timedelta(seconds=randint(60, 600))  # noqa: S311
            logger(logging.WARNING, 'RSEWriteBlocked for rule %s' % rule_id)
            METRICS.counter('exceptions.{exception}').labels(exception=e.__class__.__name__).inc()
        except ReplicationRuleCreationTemporaryFailed as e:
            paused_rules[rule_id] = datetime.utcnow() + timedelta(seconds=randint(60, 600))  # noqa: S311
            logger(logging.WARNING, 'ReplicationRuleCreationTemporaryFailed for rule %s' % rule_id)
            METRICS.counter('exceptions.{exception}').labels(exception=e.__class__.__name__).inc()
        except RuleNotFound:
            pass
        except InsufficientAccountLimit:
            # A rule with InsufficientAccountLimit on injection hangs there potentially forever
            # It should be marked as SUSPENDED
            logger(logging.INFO, 'Marking rule %s as SUSPENDED due to InsufficientAccountLimit' % rule_id)
            update_rule(rule_id=rule_id, options={'state': 'SUSPENDED'})


def stop(signum: Optional[int] = None, frame: Optional["FrameType"] = None) -> None:
    """
    Graceful exit.
    """

    graceful_stop.set()


def run(
        once: bool = False,
        threads: int = 1,
        sleep_time: int = 60
) -> None:
    """
    Starts up the Judge-Injector threads.
    """
    setup_logging(process_name=DAEMON_NAME)

    if rucio.db.sqla.util.is_old_db():
        raise DatabaseException('Database was not updated, daemon won\'t start')

    if once:
        rule_injector(once)
    else:
        logging.info('Injector starting %s threads' % str(threads))
        thread_list = [threading.Thread(target=rule_injector, kwargs={'once': once,
                                                                      'sleep_time': sleep_time}) for i in range(0, threads)]
        [t.start() for t in thread_list]
        # Interruptible joins require a timeout.
        while thread_list[0].is_alive():
            [t.join(timeout=3.14) for t in thread_list]



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import errno
import logging
import os
import signal
import subprocess
import sys
import traceback
from configparser import NoOptionError, NoSectionError
from functools import wraps

from rucio.client.client import Client
from rucio.common.config import config_get
from rucio.common.exception import (
    AccessDenied,
    CannotAuthenticate,
    DataIdentifierAlreadyExists,
    DataIdentifierNotFound,
    Duplicate,
    DuplicateContent,
    InvalidObject,
    InvalidRSEExpression,
    MissingDependency,
    RSENotFound,
    RucioException,
    RuleNotFound,
    UnsupportedOperation,
)
from rucio.common.utils import setup_logger

SUCCESS = 0
FAILURE = 1


def exception_handler(function):
    verbosity = ("-v" in sys.argv) or ("--verbose" in sys.argv)
    logger = setup_logger(module_name=__name__, logger_name="user", verbose=verbosity)

    @wraps(function)
    def new_funct(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except NotImplementedError as error:
            logger.error(f"Cannot run that operation/command combination {error}")
            return FAILURE
        except InvalidObject as error:
            logger.error(error)
            return error.error_code
        except DataIdentifierNotFound as error:
            logger.error(error)
            logger.debug("This means that the Data IDentifier you provided is not known by Rucio.")
            return error.error_code
        except AccessDenied as error:
            logger.error(error)
            logger.debug("This error is a permission issue. You cannot run this command with your account.")
            return error.error_code
        except DataIdentifierAlreadyExists as error:
            logger.error(error)
            logger.debug("This means that the Data IDentifier you try to add is already registered in Rucio.")
            return error.error_code
        except RSENotFound as error:
            logger.error(error)
            logger.debug("This means that the Rucio Storage Element you provided is not known by Rucio.")
            return error.error_code
        except InvalidRSEExpression as error:
            logger.error(error)
            logger.debug("This means the RSE expression you provided is not syntactically correct.")
            return error.error_code
        except DuplicateContent as error:
            logger.error(error)
            logger.debug("This means that the DID you want to attach is already in the target DID.")
            return error.error_code
        except Duplicate as error:
            logger.error(error)
            logger.debug("This means that you are trying to add something that already exists.")
            return error.error_code
        except TypeError as error:
            logger.error(error)
            logger.debug("This means the parameter you passed has a wrong type.")
            return FAILURE
        except RuleNotFound as error:
            logger.error(error)
            logger.debug("This means the rule you specified does not exist.")
            return error.error_code
        except UnsupportedOperation as error:
            logger.error(error)
            logger.debug("This means you cannot change the status of the DID.")
            return error.error_code
        except MissingDependency as error:
            logger.error(error)
            logger.debug("This means one dependency is missing.")
            return error.error_code
        except KeyError as error:
            if "x-rucio-auth-token" in str(error):
                used_account = None
                try:  # get the configured account from the configuration file
                    used_account = "%s (from rucio.cfg)" % config_get("client", "account")
                except Exception:
                    pass
                try:  # are we overriden by the environment?
                    used_account = "%s (from RUCIO_ACCOUNT)" % os.environ["RUCIO_ACCOUNT"]
                except Exception:
                    pass
                logger.error("Specified account %s does not have an associated identity." % used_account)

            else:
                logger.debug(traceback.format_exc())
                contact = config_get("policy", "support", raise_exception=False)
                support = ("Please follow up with all relevant information at: " + contact) if contact else ""
                logger.error("\nThe object is missing this property: %s\n" 'This should never happen. Please rerun the last command with the "-v" option to gather more information.\n' "%s" % (str(error), support))
            return FAILURE
        except RucioException as error:
            logger.error(error)
            return error.error_code
        except Exception as error:
            if isinstance(error, IOError) and getattr(error, "errno", None) == errno.EPIPE:
                # Ignore Broken Pipe
                # While in python3 we can directly catch 'BrokenPipeError', in python2 it doesn't exist.

                # Python flushes standard streams on exit; redirect remaining output
                # to devnull to avoid another BrokenPipeError at shutdown
                devnull = os.open(os.devnull, os.O_WRONLY)
                os.dup2(devnull, sys.stdout.fileno())
                return SUCCESS
            logger.debug(traceback.format_exc())
            logger.error(error)
            contact = config_get("policy", "support", raise_exception=False)
            support = ("If it's a problem concerning your experiment or if you're unsure what to do, please follow up at: %s\n" % contact) if contact else ""
            contact = config_get("policy", "support_rucio", default="https://github.com/rucio/rucio/issues")
            support += "If you're sure there is a problem with Rucio itself, please follow up at: " + contact
            logger.error("\nRucio exited with an unexpected/unknown error.\n" 'Please rerun the last command with the "-v" option to gather more information.\n' "%s" % support)
            return FAILURE

    return new_funct


def get_client(args, logger):
    """
    Returns a new client object.
    """
    if hasattr(args, "config") and (args.config is not None):
        os.environ["RUCIO_CONFIG"] = args.config

    if logger is None:
        logger = setup_logger(module_name=__name__, logger_name="user", verbose=args.verbose)

    if not args.auth_strategy:
        if "RUCIO_AUTH_TYPE" in os.environ:
            auth_type = os.environ["RUCIO_AUTH_TYPE"].lower()
        else:
            try:
                auth_type = config_get("client", "auth_type").lower()
            except (NoOptionError, NoSectionError):
                logger.error("Cannot get AUTH_TYPE")
                sys.exit(1)
    else:
        auth_type = args.auth_strategy.lower()

    if auth_type in ["userpass", "saml"] and args.username is not None and args.password is not None:
        creds = {"username": args.username, "password": args.password}
    elif auth_type == "oidc":
        if args.oidc_issuer:
            args.oidc_issuer = args.oidc_issuer.lower()
        creds = {
            "oidc_auto": args.oidc_auto,
            "oidc_scope": args.oidc_scope,
            "oidc_audience": args.oidc_audience,
            "oidc_polling": args.oidc_polling,
            "oidc_refresh_lifetime": args.oidc_refresh_lifetime,
            "oidc_issuer": args.oidc_issuer,
            "oidc_username": args.oidc_username,
            "oidc_password": args.oidc_password,
        }
    elif auth_type == "x509":
        creds = {"client_cert": args.certificate, "client_key": args.client_key}
    else:
        creds = None

    try:
        client = Client(rucio_host=args.host, auth_host=args.auth_host, account=args.issuer, auth_type=auth_type, creds=creds, ca_cert=args.ca_certificate, timeout=args.timeout, user_agent=args.user_agent, vo=args.vo, logger=logger)
    except CannotAuthenticate as error:
        logger.error(error)
        if "alert certificate expired" in str(error):
            logger.error("The server certificate expired.")
        elif auth_type.lower() == "x509_proxy":
            logger.error("Please verify that your proxy is still valid and renew it if needed.")
        sys.exit(1)
    return client


def signal_handler(sig, frame, logger):
    logger.warning("You pressed Ctrl+C! Exiting gracefully")
    child_processes = subprocess.Popen("ps -o pid --ppid %s --noheaders" % os.getpid(), shell=True, stdout=subprocess.PIPE)
    child_processes = child_processes.stdout.read()  # type: ignore
    for pid in child_processes.split("\n")[:-1]:  # type: ignore
        try:
            os.kill(int(pid), signal.SIGTERM)
        except Exception:
            print("Cannot kill child process")
    sys.exit(1)


def setup_gfal2_logger():
    gfal2_logger = logging.getLogger("gfal2")
    gfal2_logger.setLevel(logging.CRITICAL)
    gfal2_logger.addHandler(logging.StreamHandler())


class Arguments(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' add convention table and closed_at to dids '''

import datetime

import sqlalchemy as sa
from alembic import context
from alembic.op import add_column, create_check_constraint, create_foreign_key, create_primary_key, create_table, drop_column, drop_table

from rucio.common.schema import get_schema_value
from rucio.db.sqla.constants import KeyType

# Alembic revision identifiers
revision = '3082b8cef557'
down_revision = '269fee20dee9'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        add_column('dids', sa.Column('closed_at', sa.DateTime), schema=schema)
        add_column('contents_history', sa.Column('deleted_at', sa.DateTime), schema=schema)
        create_table('naming_conventions',
                     sa.Column('scope', sa.String(get_schema_value('SCOPE_LENGTH'))),
                     sa.Column('regexp', sa.String(255)),
                     sa.Column('convention_type', sa.Enum(KeyType,
                                                          name='CVT_TYPE_CHK',
                                                          create_constraint=True,
                                                          values_callable=lambda obj: [e.value for e in obj])),
                     sa.Column('updated_at', sa.DateTime, default=datetime.datetime.utcnow),
                     sa.Column('created_at', sa.DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow))
        create_primary_key('NAMING_CONVENTIONS_PK', 'naming_conventions', ['scope'])
        create_foreign_key('NAMING_CONVENTIONS_SCOPE_FK', 'naming_conventions',
                           'scopes', ['scope'], ['scope'])
        create_check_constraint('NAMING_CONVENTIONS_CREATED_NN', 'naming_conventions',
                                'created_at is not null')
        create_check_constraint('NAMING_CONVENTIONS_UPDATED_NN', 'naming_conventions',
                                'updated_at is not null')


def downgrade():
    '''
    Downgrade the database to the previous revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        drop_column('dids', 'closed_at', schema=schema)
        drop_column('contents_history', 'deleted_at', schema=schema)
        drop_table('naming_conventions', schema=schema)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
import json
import uuid
from typing import TYPE_CHECKING, Union

from flask import Blueprint, Flask, request
from werkzeug.datastructures import Headers

from rucio.core.trace import trace
from rucio.web.rest.flaskapi.v1.common import ErrorHandlingMethodView, response_headers

if TYPE_CHECKING:
    from typing import Optional

    from rucio.web.rest.flaskapi.v1.types import HeadersType


class Trace(ErrorHandlingMethodView):

    def __handle_payload_item(self, payload: dict) -> None:
        """
        Handles and processes a single trace payload item by adding various trace information.

        Args:
          payload (dict): The payload dictionary to be processed.

        Modifies:
          payload (dict): Adds the following keys to the payload:
            - 'traceTimeentry': The current UTC timestamp.
            - 'traceTimeentryUnix': The Unix timestamp with microsecond precision.
            - 'traceIp': The client's IP address, either from 'X-Forwarded-For' header or remote address.
            - 'traceId': A unique identifier for the trace, generated as a UUID without hyphens.

        Calls:
          trace(payload): A function to handle the processed payload.
        """
        # generate entry timestamp
        payload["traceTimeentry"] = datetime.datetime.now(datetime.timezone.utc)
        payload["traceTimeentryUnix"] = payload["traceTimeentry"].timestamp()

        # guess client IP
        payload["traceIp"] = request.headers.get(
            "X-Forwarded-For", default=request.remote_addr
        )

        # generate unique ID
        payload["traceId"] = str(uuid.uuid4()).replace("-", "").lower()
        trace(payload=payload)

    def get_headers(self) -> "Optional[HeadersType]":
        headers = Headers()
        headers.set('Content-Type', 'application/octet-stream')
        headers.set('Access-Control-Allow-Origin', request.environ.get('HTTP_ORIGIN'))
        headers.set('Access-Control-Allow-Headers', request.environ.get('HTTP_ACCESS_CONTROL_REQUEST_HEADERS'))
        headers.set('Access-Control-Allow-Methods', '*')
        headers.set('Access-Control-Allow-Credentials', 'true')
        return headers

    def post(self):
        """
        ---
        summary: Trace
        description: Trace endpoint used by the pilot and CLI clients to post data access information.
        tags:
          - Trace
        parameters:
        - name: X-Forwarded-For
          in: header
          schema:
            type: string
        requestBody:
          content:
            application/json:
              schema:
          oneOf:
            - type: object
              description: A single trace object.
            - type: array
              items:
                type: object
              description: A list of trace objects.
        responses:
          201:
            description: OK
          400:
            description: Cannot decode json data.
        """
        headers = self.get_headers()
        req_body: str = request.get_data(as_text=True)
        payload: Union[list, dict, None] = json.loads(req_body) if req_body else None

        if payload is None:
            return (
                "Invalid JSON data. Please provide a single trace as a JSON object or a list of trace objects.",
                400,
                headers,
            )

        if isinstance(payload, list):
            for item in payload:
                self.__handle_payload_item(item)
        else:
            self.__handle_payload_item(payload)

        return "Created", 201, headers


def blueprint():
    bp = Blueprint('traces', __name__, url_prefix='/traces')

    trace_view = Trace.as_view('trace')
    bp.add_url_rule('/', view_func=trace_view, methods=['post', ])

    bp.after_request(response_headers)
    return bp


def make_doc():
    """ Only used for sphinx documentation """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint())
    return doc_app



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import functools
import logging
import os
import queue
import socket
import threading
import time
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar, Union

from rucio.common.logging import formatted_logger
from rucio.common.utils import PriorityQueue
from rucio.core import heartbeat as heartbeat_core
from rucio.core.monitor import MetricManager

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Iterator, Sequence

    from rucio.common.types import LoggerFunction

T = TypeVar('T')
METRICS = MetricManager(module=__name__)


class HeartbeatHandler:
    """
    Simple contextmanager which sets a heartbeat and associated logger on entry and cleans up the heartbeat on exit.
    """

    def __init__(self, executable: str, renewal_interval: int):
        """
        :param executable: the executable name which will be set in heartbeats
        :param renewal_interval: the interval at which the heartbeat will be renewed in the database.
        Calls to live() in-between intervals will reuse the locally cached heartbeat.
        """
        self.executable = executable
        self._hash_executable = None
        self.renewal_interval = renewal_interval
        self.older_than = renewal_interval * 10 if renewal_interval and renewal_interval > 0 else None  # 10 was chosen without any particular reason

        self.hostname = socket.getfqdn()
        self.pid = os.getpid()
        self.hb_thread = threading.current_thread()

        self.logger = logging.log
        self.last_heart_beat = None
        self.last_time = None
        self.last_payload = None

    def __enter__(self) -> 'HeartbeatHandler':
        heartbeat_core.sanity_check(executable=self.executable, hostname=self.hostname)
        self.live()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self.last_heart_beat:
            heartbeat_core.die(self.executable, self.hostname, self.pid, self.hb_thread)
            if self.logger:
                self.logger(logging.INFO, 'Heartbeat cleaned up')

    @property
    def hash_executable(self) -> str:
        if not self._hash_executable:
            self._hash_executable = heartbeat_core.calc_hash(self.executable)
        return self._hash_executable

    @property
    def short_executable(self) -> str:
        return min(self.executable, self.hash_executable, key=len)

    def live(
            self,
            force_renew: bool = False,
            payload: Optional[str] = None
    ) -> tuple[int, int, 'Callable']:
        """
        :return: a tuple: <the number of the current worker>, <total number of workers>, <decorated logger>
        """
        if force_renew \
                or not self.last_time \
                or not self.last_heart_beat \
                or self.last_time < datetime.datetime.now() - datetime.timedelta(seconds=self.renewal_interval) \
                or self.last_payload != payload:
            if self.older_than:
                self.last_heart_beat = heartbeat_core.live(self.executable, self.hostname, self.pid, self.hb_thread, payload=payload, older_than=self.older_than)
            else:
                self.last_heart_beat = heartbeat_core.live(self.executable, self.hostname, self.pid, self.hb_thread, payload=payload)

            prefix = '[%i/%i]: ' % (self.last_heart_beat['assign_thread'], self.last_heart_beat['nr_threads'])
            self.logger = formatted_logger(logging.log, prefix + '%s')

            if not self.last_time:
                self.logger(logging.DEBUG, 'First heartbeat set')
            else:
                self.logger(logging.DEBUG, 'Heartbeat renewed')
            self.last_time = datetime.datetime.now()
            self.last_payload = payload

        return self.last_heart_beat['assign_thread'], self.last_heart_beat['nr_threads'], self.logger


def _activity_looper(
        once: bool,
        sleep_time: int,
        activities: Optional['Sequence[str]'],
        heartbeat_handler: HeartbeatHandler,
) -> 'Generator[tuple[str, float], tuple[float, bool], None]':
    """
    Generator which loops (either once, or indefinitely) over all activities while ensuring that `sleep_time`
    passes between handling twice the same activity.

    Returns an activity and how much time the calling context must sleep before handling that activity
    and expects to get in return the time when the activity started to be executed and whether next
    execution must be immediate.
    """

    # For each activity, the priority queue will keep the next absolute time when that
    # activity must be handled.
    activity_next_exe_time = PriorityQueue()

    # On startup, we schedule to immediately handle all activities.
    now = time.time()
    for activity in activities or [None]:
        activity_next_exe_time[activity] = now

    while activity_next_exe_time:
        activity = activity_next_exe_time.top()
        desired_exe_time = activity_next_exe_time[activity]

        if once:
            time_to_sleep = 0
            activity_next_exe_time.pop()
        else:
            time_to_sleep = desired_exe_time - time.time()

        logger = heartbeat_handler.logger
        if time_to_sleep > 0:
            if activity:
                logger(logging.DEBUG, 'Switching to activity %s and sleeping %s seconds', activity, time_to_sleep)
            else:
                logger(logging.DEBUG, 'Sleeping %s seconds', time_to_sleep)
        else:
            if activity:
                logger(logging.DEBUG, 'Switching to activity %s', activity)
            else:
                logger(logging.DEBUG, 'Starting next iteration')

        # The calling context notifies us when the activity actually got handled. And if sleeping is desired.
        actual_exe_time, must_sleep = yield activity, time_to_sleep

        if not once:
            if must_sleep:
                time_diff = time.time() - actual_exe_time
                time_to_sleep = max(1.0, sleep_time - time_diff)
                activity_next_exe_time[activity] = time.time() + time_to_sleep
            else:
                activity_next_exe_time[activity] = time.time() + 1


def db_workqueue(
        once: bool,
        graceful_stop: threading.Event,
        executable: str,
        partition_wait_time: int,
        sleep_time: int,
        activities: Optional['Sequence[str]'] = None,
) -> 'Callable[[Callable[..., Union[bool, tuple[bool, T], None]]], Callable[[], Iterator[Union[T, None]]]]':
    """
    Used to wrap a function for interacting with the database as a work queue: i.e. to select
    a set of rows and perform some work on those rows while ensuring that two instances running in parallel don't
    work on the same set of rows. The last condition is ensured by using heartbeats to keep track of currently
    active workers.

    :param once: Whether to stop after one iteration
    :param graceful_stop: the threading.Event() object used for graceful stop of the daemon
    :param executable: the name of the executable used for heartbeats
    :param partition_wait_time: time to wait for database partition rebalancing before starting the actual daemon loop
    :param sleep_time: time to sleep between the iterations of the daemon
    :param activities: optional list of activities on which to work. The run_once_fnc will be called on activities one by one.
    """

    def _decorate(run_once_fnc: 'Callable[..., Optional[Union[bool, tuple[bool, T]]]]') -> 'Callable[[], Iterator[Optional[T]]]':

        @functools.wraps(run_once_fnc)
        def _generator() -> 'Iterator[T]':

            with HeartbeatHandler(executable=executable, renewal_interval=sleep_time - 1) as heartbeat_handler:
                logger = heartbeat_handler.logger
                logger(logging.INFO, 'started')

                if partition_wait_time:
                    graceful_stop.wait(partition_wait_time)
                    _, _, logger = heartbeat_handler.live(force_renew=True)

                activity_loop = _activity_looper(once=once, sleep_time=sleep_time, activities=activities, heartbeat_handler=heartbeat_handler)
                activity, time_to_sleep = next(activity_loop, (None, None))
                while time_to_sleep is not None:
                    if graceful_stop.is_set():
                        break

                    if time_to_sleep > 0:
                        graceful_stop.wait(time_to_sleep)

                    _, _, logger = heartbeat_handler.live()

                    must_sleep = True
                    start_time = time.time()
                    try:
                        result = run_once_fnc(heartbeat_handler=heartbeat_handler, activity=activity)

                        # Handle return values already existing in the code
                        # TODO: update all existing daemons to always explicitly return (must_sleep, ret_value)
                        if result is None:
                            must_sleep = True
                            ret_value = None
                        elif isinstance(result, bool):
                            must_sleep = result
                            ret_value = None
                        else:
                            must_sleep, ret_value = result

                        if ret_value is not None:
                            yield ret_value
                    except Exception as e:
                        METRICS.counter('exceptions.{exception}').labels(exception=e.__class__.__name__).inc()
                        logger(logging.CRITICAL, "Exception", exc_info=True)
                        if once:
                            raise

                    try:
                        activity, time_to_sleep = activity_loop.send((start_time, must_sleep))
                    except StopIteration:
                        break

                if not once:
                    logger(logging.INFO, 'Graceful stop requested')

        return _generator

    return _decorate


def run_daemon(
        once: bool,
        graceful_stop: threading.Event,
        executable: str,
        partition_wait_time: int,
        sleep_time: int,
        run_once_fnc: 'Callable[..., Optional[Union[bool, tuple[bool, Any]]]]',
        activities: Optional[list[str]] = None
) -> None:
    """
    Run the daemon loop and call the function run_once_fnc at each iteration
    """

    daemon = db_workqueue(
        once=once,
        graceful_stop=graceful_stop,
        executable=executable,
        partition_wait_time=partition_wait_time,
        sleep_time=sleep_time,
        activities=activities,
    )(run_once_fnc)

    for _ in daemon():
        pass


class ProducerConsumerDaemon(Generic[T]):
    """
    Daemon which connects N producers with M consumers via a queue.
    """

    def __init__(
            self,
            producers: 'Sequence[Callable[[], Iterator[T]]]',
            consumers: 'Sequence[Callable[..., None]]',
            graceful_stop: threading.Event,
            logger: "LoggerFunction" = logging.log
    ):
        self.producers = producers
        self.consumers = consumers

        self.queue = queue.Queue()
        self.lock = threading.Lock()
        self.graceful_stop = graceful_stop
        self.active_producers = 0
        self.producers_done_event = threading.Event()
        self.logger = logger

    def _produce(
            self,
            it: 'Callable[[], Iterator[T]]',
            wait_for_consumers: bool = False
    ) -> None:
        """
        Iterate over the generator function and put the extracted elements into the queue.

        Perform a graceful shutdown when graceful_stop is set.
        """

        i = it()
        with self.lock:
            self.active_producers += 1
        try:
            while not self.graceful_stop.is_set():
                if self.queue.qsize() > len(self.consumers):
                    self.graceful_stop.wait(1)
                    continue

                try:
                    product = next(i)
                    self.queue.put(product)
                except StopIteration:
                    break
                except Exception as e:
                    METRICS.counter('exceptions.{exception}').labels(exception=e.__class__.__name__).inc()
                    self.logger(logging.CRITICAL, "Exception", exc_info=True)
        finally:
            with self.lock:
                self.active_producers -= 1
                if not self.active_producers > 0:
                    self.producers_done_event.set()

            if wait_for_consumers:
                self.queue.join()

    def _consume(
            self,
            fnc: 'Callable[[T], Any]'
    ) -> None:
        """
        Wait for elements to arrive via the queue and call the given function on each element.

        If producers_done_event is set, handle all remaining elements from the queue and exit gracefully.
        """
        while not self.producers_done_event.is_set() or self.queue.unfinished_tasks:
            try:
                product = self.queue.get_nowait()
            except queue.Empty:
                self.producers_done_event.wait(1)
                continue

            try:
                fnc(product)
            except Exception as e:
                METRICS.counter('exceptions.{exception}').labels(exception=e.__class__.__name__).inc()
                self.logger(logging.CRITICAL, "Exception", exc_info=True)
            finally:
                self.queue.task_done()

    def run(self) -> None:

        producer_threads: list[threading.Thread] = []
        for i, producer in enumerate(self.producers):
            thread = threading.Thread(
                target=self._produce,
                name=f'producer-{i}-{producer.__name__}',
                kwargs={
                    'it': producer,
                    'wait_for_consumers': True
                }
            )
            thread.start()
            producer_threads.append(thread)

        consumer_threads: list[threading.Thread] = []
        for i, consumer in enumerate(self.consumers):
            thread = threading.Thread(
                target=self._consume,
                name=f'consumer-{i}-{consumer.__name__}',
                kwargs={
                    'fnc': consumer,
                }
            )
            thread.start()
            consumer_threads.append(thread)

        logging.info('waiting for interrupts')

        while producer_threads:
            for thread in producer_threads:
                thread.join(timeout=3.14)
            producer_threads = [thread for thread in producer_threads if thread.is_alive()]

        self.producers_done_event.set()

        while consumer_threads:
            for thread in consumer_threads:
                thread.join(timeout=3.14)
            consumer_threads = [thread for thread in consumer_threads if thread.is_alive()]



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import TYPE_CHECKING

from flask import Blueprint, Flask, request
from werkzeug.datastructures import Headers

from rucio.core.nongrid_trace import trace
from rucio.web.rest.flaskapi.v1.common import ErrorHandlingMethodView, json_parameters, response_headers

if TYPE_CHECKING:
    from typing import Optional

    from rucio.web.rest.flaskapi.v1.types import HeadersType


class XAODTrace(ErrorHandlingMethodView):

    def get_headers(self) -> "Optional[HeadersType]":
        headers = Headers()
        headers.set('Content-Type', 'application/octet-stream')
        headers.set('Access-Control-Allow-Origin', request.environ.get('HTTP_ORIGIN'))
        headers.set('Access-Control-Allow-Headers', request.environ.get('HTTP_ACCESS_CONTROL_REQUEST_HEADERS'))
        headers.set('Access-Control-Allow-Methods', '*')
        headers.set('Access-Control-Allow-Credentials', 'true')
        return headers

    def post(self):
        """
        ---
        summary: Trace endpoints
        description: Trace endpoint used by the XAOD framework to post data access information.
        tags:
          - Nongrid traces
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  payload:
                    description: Dictionary containing the trace information.
                    type: object
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Created"]
          400:
            description: Cannot decode json data.
        """
        headers = self.get_headers()

        parameters = json_parameters()

        # generate entry timestamp
        parameters['timeentry'] = int(time.time())

        # guess client IP
        parameters['ip'] = request.headers.get('X-Forwarded-For', default=request.remote_addr)

        trace(payload=parameters)

        return 'Created', 201, headers


def blueprint():
    bp = Blueprint('nongrid_traces', __name__, url_prefix='/nongrid_traces')

    xaod_trace_view = XAODTrace.as_view('xaod_trace')
    bp.add_url_rule('/', view_func=xaod_trace_view, methods=['post', ])

    bp.after_request(response_headers)
    return bp


def make_doc():
    """ Only used for sphinx documentation """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint())
    return doc_app



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from sqlalchemy import inspect, update
from sqlalchemy.exc import CompileError, InvalidRequestError, NoResultFound
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import true

from rucio.common import exception
from rucio.core import account_counter, rse_counter
from rucio.core.did_meta_plugins.did_meta_plugin_interface import DidMetaPlugin
from rucio.core.did_meta_plugins.filter_engine import FilterEngine
from rucio.db.sqla import models
from rucio.db.sqla.constants import DIDType
from rucio.db.sqla.session import read_session, stream_session, transactional_session

if TYPE_CHECKING:
    from typing import Optional

    from sqlalchemy.orm import Session


class DidColumnMeta(DidMetaPlugin):
    """
    A metadata plugin to interact with the base did table metadata.
    """
    def __init__(self):
        super(DidColumnMeta, self).__init__()
        self.plugin_name = "DID_COLUMN"

    @read_session
    def get_metadata(self, scope, name, *, session: "Session"):
        """
        Get data identifier metadata.

        :param scope: The scope name.
        :param name: The data identifier name.
        :param session: The database session in use.
        """
        try:
            row = session.query(models.DataIdentifier).filter_by(scope=scope, name=name).\
                with_hint(models.DataIdentifier, "INDEX(DIDS DIDS_PK)", 'oracle').one()
            return row.to_dict()
        except NoResultFound:
            raise exception.DataIdentifierNotFound(f"Data identifier '{scope}:{name}' not found")

    @transactional_session
    def set_metadata(self, scope, name, key, value, recursive=False, *, session: "Session"):
        self.set_metadata_bulk(scope=scope, name=name, metadata={key: value}, recursive=recursive, session=session)

    @transactional_session
    def set_metadata_bulk(self, scope, name, metadata, recursive=False, *, session: "Session"):
        did_query = session.query(models.DataIdentifier).with_hint(models.DataIdentifier, "INDEX(DIDS DIDS_PK)", 'oracle').filter_by(scope=scope, name=name)
        if did_query.one_or_none() is None:
            raise exception.DataIdentifierNotFound("Data identifier '%s:%s' not found" % (scope, name))

        remainder = {}
        for key, value in metadata.items():
            if key == 'eol_at' and isinstance(value, str):
                try:
                    eol_at = datetime.strptime(value, '%Y-%M-%d')
                    rowcount = did_query.update({'eol_at': eol_at}, synchronize_session='fetch')
                except TypeError as error:
                    raise exception.InvalidValueForKey(error)
                if not rowcount:
                    # check for did presence
                    raise exception.UnsupportedOperation('%s for %s:%s cannot be updated' % (key, scope, name))
            elif key == 'lifetime':
                try:
                    expired_at = None
                    if value is not None:
                        expired_at = datetime.utcnow() + timedelta(seconds=float(value))
                    rowcount = did_query.update({'expired_at': expired_at}, synchronize_session='fetch')
                except TypeError as error:
                    raise exception.InvalidValueForKey(error)
                if not rowcount:
                    # check for did presence
                    raise exception.UnsupportedOperation('%s for %s:%s cannot be updated' % (key, scope, name))
            elif key in ['guid', 'events']:
                rowcount = did_query.filter_by(did_type=DIDType.FILE).update({key: value}, synchronize_session=False)
                if not rowcount:
                    # check for did presence
                    raise exception.UnsupportedOperation('%s for %s:%s cannot be updated' % (key, scope, name))

                session.query(models.DataIdentifierAssociation).filter_by(child_scope=scope, child_name=name, child_type=DIDType.FILE).update({key: value}, synchronize_session=False)
                if key == 'events':
                    for parent_scope, parent_name in session.query(models.DataIdentifierAssociation.scope, models.DataIdentifierAssociation.name).filter_by(child_scope=scope, child_name=name):
                        events = session.query(func.sum(models.DataIdentifierAssociation.events)).filter_by(scope=parent_scope, name=parent_name).one()[0]
                        session.query(models.DataIdentifier).filter_by(scope=parent_scope, name=parent_name).update({'events': events}, synchronize_session=False)
            elif key == 'adler32':
                rowcount = did_query.filter_by(did_type=DIDType.FILE).update({key: value}, synchronize_session=False)
                if not rowcount:
                    # check for did presence
                    raise exception.UnsupportedOperation('%s for %s:%s cannot be updated' % (key, scope, name))

                session.query(models.DataIdentifierAssociation).filter_by(child_scope=scope, child_name=name, child_type=DIDType.FILE).update({key: value}, synchronize_session=False)
                session.query(models.Request).filter_by(scope=scope, name=name).update({key: value}, synchronize_session=False)
                session.query(models.RSEFileAssociation).filter_by(scope=scope, name=name).update({key: value}, synchronize_session=False)
            elif key == 'bytes':
                rowcount = did_query.filter_by(did_type=DIDType.FILE).update({key: value}, synchronize_session=False)
                if not rowcount:
                    # check for did presence
                    raise exception.UnsupportedOperation('%s for %s:%s cannot be updated' % (key, scope, name))

                session.query(models.DataIdentifierAssociation).filter_by(child_scope=scope, child_name=name, child_type=DIDType.FILE).update({key: value}, synchronize_session=False)
                session.query(models.Request).filter_by(scope=scope, name=name).update({key: value}, synchronize_session=False)

                for account, bytes_, rse_id, rule_id in session.query(models.ReplicaLock.account, models.ReplicaLock.bytes, models.ReplicaLock.rse_id, models.ReplicaLock.rule_id).filter_by(scope=scope, name=name):
                    session.query(models.ReplicaLock).filter_by(scope=scope, name=name, rule_id=rule_id, rse_id=rse_id).update({key: value}, synchronize_session=False)
                    account_counter.decrease(rse_id=rse_id, account=account, files=1, bytes_=bytes_, session=session)
                    account_counter.increase(rse_id=rse_id, account=account, files=1, bytes_=value, session=session)

                for bytes_, rse_id in session.query(models.RSEFileAssociation.bytes, models.RSEFileAssociation.rse_id).filter_by(scope=scope, name=name):
                    session.query(models.RSEFileAssociation).filter_by(scope=scope, name=name, rse_id=rse_id).update({key: value}, synchronize_session=False)
                    rse_counter.decrease(rse_id=rse_id, files=1, bytes_=bytes_, session=session)
                    rse_counter.increase(rse_id=rse_id, files=1, bytes_=value, session=session)

                for parent_scope, parent_name in session.query(models.DataIdentifierAssociation.scope, models.DataIdentifierAssociation.name).filter_by(child_scope=scope, child_name=name):
                    values = {}
                    values['length'], values['bytes'], values['events'] = session.query(func.count(models.DataIdentifierAssociation.scope),
                                                                                        func.sum(models.DataIdentifierAssociation.bytes),
                                                                                        func.sum(models.DataIdentifierAssociation.events)).filter_by(scope=parent_scope, name=parent_name).one()
                    session.query(models.DataIdentifier).filter_by(scope=parent_scope, name=parent_name).update(values, synchronize_session=False)
                    session.query(models.DatasetLock).filter_by(scope=parent_scope, name=parent_name).update({'length': values['length'], 'bytes': values['bytes']}, synchronize_session=False)
            else:
                remainder[key] = value

        if remainder:
            try:
                rowcount = did_query.update(remainder, synchronize_session='fetch')
            except CompileError as error:
                raise exception.InvalidMetadata(error)
            except InvalidRequestError:
                raise exception.InvalidMetadata("Some of the keys are not accepted: " + str(list(remainder.keys())))
            if not rowcount:
                raise exception.UnsupportedOperation('Some of the keys for %s:%s cannot be updated: %s' % (scope, name, str(list(remainder.keys()))))

            # propagate metadata updates to child content
            if recursive:
                content_query = session.query(models.DataIdentifierAssociation.child_scope, models.DataIdentifierAssociation.child_name)
                content_query = content_query.with_hint(models.DataIdentifierAssociation, "INDEX(CONTENTS CONTENTS_PK)", 'oracle').filter_by(scope=scope, name=name)

                for child_scope, child_name in content_query:
                    try:
                        stmt = update(models.DataIdentifier)\
                            .prefix_with("/*+ INDEX(DIDS DIDS_PK) */", dialect='oracle')\
                            .filter_by(scope=child_scope, name=child_name)\
                            .execution_options(synchronize_session='fetch')\
                            .values(remainder)
                        session.execute(stmt)
                    except CompileError as error:
                        raise exception.InvalidMetadata(error)
                    except InvalidRequestError:
                        raise exception.InvalidMetadata("Some of the keys are not accepted recursively: " + str(list(remainder.keys())))

    @stream_session
    def list_dids(self, scope, filters, did_type='collection', ignore_case=False, limit=None,
                  offset=None, long=False, recursive=False, ignore_dids=None, *, session: "Session"):
        """
        Search data identifiers.

        :param scope: the scope name.
        :param filters: dictionary of attributes by which the results should be filtered.
        :param did_type: the type of the did: all(container, dataset, file), collection(dataset or container), dataset, container, file.
        :param ignore_case: ignore case distinctions.
        :param limit: limit number.
        :param offset: offset number.
        :param long: Long format option to display more information for each DID.
        :param session: The database session in use.
        :param recursive: Recursively list DIDs content.
        :param ignore_dids: List of DIDs to refrain from yielding.
        """
        if not ignore_dids:
            ignore_dids = set()

        # mapping for semantic <type> to a (set of) recognised DIDType(s).
        type_to_did_type_mapping = {
            'all': [DIDType.CONTAINER, DIDType.DATASET, DIDType.FILE],
            'collection': [DIDType.CONTAINER, DIDType.DATASET],
            'container': [DIDType.CONTAINER],
            'dataset': [DIDType.DATASET],
            'file': [DIDType.FILE]
        }

        # backwards compatibility for filters as single {}.
        if isinstance(filters, dict):
            filters = [filters]

        # for each or_group, make sure there is a mapped "did_type" filter.
        # if type maps to many DIDTypes, the corresponding or_group will be copied the required number of times to satisfy all the logical possibilities.
        filters_tmp = []
        for or_group in filters:
            if 'type' not in or_group:
                or_group_type = did_type.lower()
            else:
                or_group_type = or_group.pop('type').lower()
            if or_group_type not in type_to_did_type_mapping.keys():
                raise exception.UnsupportedOperation('{} is not a valid type. Valid types are {}'.format(or_group_type, type_to_did_type_mapping.keys()))

            for mapped_did_type in type_to_did_type_mapping[or_group_type]:
                or_group['did_type'] = mapped_did_type
                filters_tmp.append(or_group.copy())
        filters = filters_tmp

        # instantiate fe and create sqla query
        fe = FilterEngine(filters, model_class=models.DataIdentifier)
        stmt = fe.create_sqla_query(
            additional_model_attributes=[
                models.DataIdentifier.scope,
                models.DataIdentifier.name,
                models.DataIdentifier.did_type,
                models.DataIdentifier.bytes,
                models.DataIdentifier.length
            ], additional_filters=[
                (models.DataIdentifier.scope, operator.eq, scope),
                (models.DataIdentifier.suppressed, operator.ne, true())
            ],
            session=session
        )
        stmt = stmt.with_hint(
            models.DataIdentifier,
            'NO_EXPAND',
            'oracle'
        )

        if limit:
            stmt = stmt.limit(
                limit
            )
        if recursive:
            from rucio.core.did import list_content

            # Get attached DIDs and save in list because query has to be finished before starting a new one in the recursion
            collections_content = []
            for did in session.execute(stmt).yield_per(100):
                if (did.did_type == DIDType.CONTAINER or did.did_type == DIDType.DATASET):
                    collections_content += [d for d in list_content(scope=did.scope, name=did.name)]

            # Replace any name filtering with recursed DID names.
            for did in collections_content:
                for or_group in filters:
                    or_group['name'] = did['name']
                for result in self.list_dids(scope=did['scope'], filters=filters, recursive=True, did_type=did_type, limit=limit, offset=offset,
                                             long=long, ignore_dids=ignore_dids, session=session):
                    yield result

        for did in session.execute(stmt).yield_per(5):                  # don't unpack this as it makes it dependent on query return order!
            if long:
                did_full = "{}:{}".format(did.scope, did.name)
                if did_full not in ignore_dids:         # concatenating results of OR clauses may contain duplicate DIDs if query result sets not mutually exclusive.
                    ignore_dids.add(did_full)
                    yield {
                        'scope': did.scope,
                        'name': did.name,
                        'did_type': did.did_type.name,
                        'bytes': did.bytes,
                        'length': did.length
                    }
            else:
                did_full = "{}:{}".format(did.scope, did.name)
                if did_full not in ignore_dids:         # concatenating results of OR clauses may contain duplicate DIDs if query result sets not mutually exclusive.
                    ignore_dids.add(did_full)
                    yield did.name

    def delete_metadata(self, scope, name, key, *, session: "Optional[Session]" = None):
        """
        Deletes the metadata stored for the given key.

        :param scope: The scope of the did.
        :param name: The name of the did.
        :param key: Key of the metadata.
        """
        raise NotImplementedError('The DidColumnMeta plugin does not currently support deleting metadata.')

    def manages_key(self, key, *, session: "Optional[Session]" = None):
        # Build list of which keys are managed by this plugin.
        #
        all_did_table_columns = []
        for column in inspect(models.DataIdentifier).attrs:
            all_did_table_columns.append(column.key)

        exclude_did_table_columns = [
            'account',
            'availability',
            'complete',
            'created_at',
            'did_type',
            'is_open',
            'monotonic',
            'obsolete',
            'scope',
            'suppressed',
            'updated_at'
        ]

        additional_keys = [
            'lifetime',
            'created_before',
            'created_after',
            'length.gt',
            'length.lt',
            'length.gte',
            'length.lte',
            'type'
        ]

        hardcoded_keys = list(set(all_did_table_columns) - set(exclude_did_table_columns)) + additional_keys

        return key in hardcoded_keys

    def get_plugin_name(self):
        """
        Returns a unique identifier for this plugin. This can be later used for filtering down results to this plugin only.
        :returns: The name of the plugin.
        """
        return self.plugin_name



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' remove temporary dids '''

import sqlalchemy as sa
from alembic import context
from alembic.op import create_index, create_primary_key, create_table, drop_table

from rucio.common.schema import get_schema_value
from rucio.db.sqla.types import GUID, InternalScopeString

# Alembic revision identifiers
revision = '4df2c5ddabc0'
down_revision = '27e3a68927fb'


def upgrade():
    '''Upgrade the database to this revision'''
    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        drop_table('tmp_dids')


def downgrade():
    '''Downgrade the database to the previous revision'''
    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        create_table('tmp_dids',
                     sa.Column('scope', InternalScopeString(get_schema_value('SCOPE_LENGTH'))),
                     sa.Column('name', sa.String(get_schema_value('NAME_LENGTH'))),
                     sa.Column('rse_id', GUID()),
                     sa.Column('path', sa.String(1024)),
                     sa.Column('bytes', sa.BigInteger),
                     sa.Column('md5', sa.String(32)),
                     sa.Column('adler32', sa.String(8)),
                     sa.Column('expired_at', sa.DateTime),
                     sa.Column('guid', GUID()),
                     sa.Column('events', sa.BigInteger),
                     sa.Column('task_id', sa.Integer),
                     sa.Column('panda_id', sa.Integer),
                     sa.Column('parent_scope', InternalScopeString(get_schema_value('SCOPE_LENGTH'))),
                     sa.Column('parent_name', sa.String(get_schema_value('NAME_LENGTH'))),
                     sa.Column('offset', sa.BigInteger))
        create_primary_key('TMP_DIDS_PK', 'tmp_dids', ['scope', 'name'])
        create_index('TMP_DIDS_EXPIRED_AT_IDX', 'tmp_dids', ['expired_at'])



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import importlib
import logging
from configparser import NoOptionError, NoSectionError
from os import environ
from typing import TYPE_CHECKING, Any

from jsonschema import ValidationError, validate

from rucio.common import config, exception
from rucio.common.plugins import check_policy_package_version

if TYPE_CHECKING:
    from types import ModuleType

LOGGER = logging.getLogger('policy')

# dictionary of schema modules for each VO
schema_modules: dict[str, "ModuleType"] = {}

# list of unique SCOPE_NAME_REGEXP values from all schemas
scope_name_regexps: list[str] = []


# cached function to check for multivo
@functools.cache
def _is_multivo():
    try:
        return config.config_get_bool('common', 'multi_vo', check_config_table=False)
    except (NoOptionError, NoSectionError):
        return False


# cached function to get generic schema module
@functools.cache
def _get_generic_schema_module():
    generic_fallback = 'generic_multi_vo' if _is_multivo() else 'generic'
    return importlib.import_module('rucio.common.schema.' + generic_fallback)


# multi-VO version loads schema per-VO on demand
# we can't get a list of VOs here because the database might not
# be available as this is imported during the bootstrapping process
if not _is_multivo():
    GENERIC_FALLBACK = 'generic'

    if config.config_has_section('policy'):
        try:
            if 'RUCIO_POLICY_PACKAGE' in environ:
                policy = environ['RUCIO_POLICY_PACKAGE']
            else:
                policy = config.config_get('policy', 'package', check_config_table=False)
            check_policy_package_version(policy)
            policy = policy + ".schema"
        except (NoOptionError, NoSectionError):
            # fall back to old system for now
            try:
                policy = config.config_get('policy', 'schema', check_config_table=False)
            except (NoOptionError, NoSectionError):
                policy = GENERIC_FALLBACK
            policy = 'rucio.common.schema.' + policy.lower()
    else:
        policy = 'rucio.common.schema.' + GENERIC_FALLBACK.lower()

    try:
        module = importlib.import_module(policy)
    except ModuleNotFoundError:
        # if policy package does not contain schema module, load fallback module instead
        # this allows a policy package to omit modules that do not need customisation
        try:
            LOGGER.warning('Unable to load schema module %s from policy package, falling back to %s'
                           % (policy, GENERIC_FALLBACK))
            policy = 'rucio.common.schema.' + GENERIC_FALLBACK.lower()
            module = importlib.import_module(policy)
        except ModuleNotFoundError:
            raise exception.PolicyPackageNotFound(policy)
        except ImportError:
            raise exception.ErrorLoadingPolicyPackage(policy)
    except ImportError:
        raise exception.ErrorLoadingPolicyPackage(policy)

    schema_modules["def"] = module
    if hasattr(module, 'SCOPE_NAME_REGEXP'):
        scope_name_regexps.append(module.SCOPE_NAME_REGEXP)


def load_schema_for_vo(vo: str) -> None:
    generic_fallback = 'generic_multi_vo'
    if config.config_has_section('policy'):
        try:
            env_name = 'RUCIO_POLICY_PACKAGE_' + vo.upper()
            if env_name in environ:
                policy = environ[env_name]
            else:
                policy = config.config_get('policy', 'package-' + vo, check_config_table=False)
            check_policy_package_version(policy)
            policy = policy + ".schema"
        except (NoOptionError, NoSectionError):
            # fall back to old system for now
            try:
                policy = config.config_get('policy', 'schema', check_config_table=False)
            except (NoOptionError, NoSectionError):
                policy = generic_fallback
            policy = 'rucio.common.schema.' + policy.lower()
    else:
        policy = 'rucio.common.schema.' + generic_fallback.lower()

    try:
        module = importlib.import_module(policy)
    except ModuleNotFoundError:
        # if policy package does not contain schema module, load fallback module instead
        # this allows a policy package to omit modules that do not need customisation
        try:
            LOGGER.warning('Unable to load schema module %s from policy package, falling back to %s'
                           % (policy, generic_fallback))
            policy = 'rucio.common.schema.' + generic_fallback.lower()
            module = importlib.import_module(policy)
        except ModuleNotFoundError:
            raise exception.PolicyPackageNotFound(policy)
        except ImportError:
            raise exception.ErrorLoadingPolicyPackage(policy)
    except ImportError:
        raise exception.ErrorLoadingPolicyPackage(policy)

    schema_modules[vo] = module


def validate_schema(name: str, obj: Any, vo: str = 'def') -> None:
    if obj:
        if vo not in schema_modules:
            load_schema_for_vo(vo)
        if hasattr(schema_modules[vo], 'SCHEMAS') and name in schema_modules[vo].SCHEMAS:
            schema = schema_modules[vo].SCHEMAS.get(name, {})
        else:
            # if schema not available in VO module, fall back to generic module
            schema = _get_generic_schema_module().SCHEMAS.get(name, {})
        try:
            validate(obj, schema)
        except ValidationError as error:  # NOQA: F841
            raise exception.InvalidObject(f'Problem validating {name}: {error}')


def get_schema_value(key: str, vo: str = 'def') -> Any:
    if vo not in schema_modules:
        load_schema_for_vo(vo)
    if not hasattr(schema_modules[vo], key):
        return getattr(_get_generic_schema_module(), key)
    return getattr(schema_modules[vo], key)


def get_scope_name_regexps() -> list[str]:
    """ returns a list of all unique SCOPE_NAME_REGEXPs from all schemas """

    if len(scope_name_regexps) == 0:
        # load schemas for all VOs here and add unique scope_name_regexps to list
        from rucio.core.vo import list_vos
        vos = list_vos()
        for vo in vos:
            if vo['vo'] not in schema_modules:
                load_schema_for_vo(vo['vo'])
            scope_name_regexp = schema_modules[vo['vo']].SCOPE_NAME_REGEXP
            if scope_name_regexp not in scope_name_regexps:
                scope_name_regexps.append(scope_name_regexp)
    return scope_name_regexps


def insert_scope_name(urls: tuple[str, ...]) -> tuple[str, str]:
    """
    given a tuple of URLs for webpy with '%s' as a placeholder for
    SCOPE_NAME_REGEXP, return a finalised tuple of URLs that will work for all
    SCOPE_NAME_REGEXPs in all schemas
    """

    regexps = get_scope_name_regexps()
    result = []
    for i in range(0, len(urls), 2):
        if "%s" in urls[i]:
            # add a copy for each unique SCOPE_NAME_REGEXP
            for scope_name_regexp in regexps:
                result.append(urls[i] % scope_name_regexp)
                result.append(urls[i + 1])
        else:
            # pass through unmodified
            result.append(urls[i])
            result.append(urls[i + 1])
    return tuple(result)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import click

from rucio.cli.bin_legacy.rucio import list_rses
from rucio.cli.bin_legacy.rucio_admin import (
    add_distance_rses,
    add_protocol_rse,
    add_qos_policy,
    add_rse,
    del_protocol_rse,
    delete_attribute_rse,
    delete_distance_rses,
    delete_limit_rse,
    delete_qos_policy,
    disable_rse,
    get_attribute_rse,
    get_distance_rses,
    info_rse,
    list_qos_policies,
    set_attribute_rse,
    set_limit_rse,
    update_distance_rses,
    update_rse,
)
from rucio.cli.utils import Arguments


@click.group()
def rse():
    """Manage Rucio Storage Elements (RSEs)"""


@rse.command("list")
@click.option("--rses", "--rse-exp", help="RSE Expression to use as a filter", required=False)
@click.option("--csv", is_flag=True, default=False, help="Output list of RSEs as a csv")
@click.pass_context
def list_(ctx, rses, csv):
    """List all registered Rucio Storage Elements (RSEs)"""
    list_rses(Arguments({"rses": rses, "csv": csv}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@rse.command("show")
@click.argument("rse-name")
@click.option("--csv", is_flag=True, default=False, help="Output list of RSE property key and values as a csv")
@click.pass_context
def show(ctx, rse_name, csv):
    """Usage, protocols, settings, and attributes for a given RSE"""
    info_rse(Arguments({"rse": rse_name, "csv": csv}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@rse.command("add")
@click.argument("rse-name")
@click.option("--non-deterministic", is_flag=True, default=False, help="Create RSE in non-deterministic mode")
@click.pass_context
def add_(ctx, rse_name, non_deterministic):
    """Add a new RSE"""
    args = Arguments({"rse": rse_name, "non_deterministic": non_deterministic})
    add_rse(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@rse.command("remove")
@click.argument("rse-name")
@click.pass_context
def remove(ctx, rse_name):
    """Permanently disable an RSE. CAUTION: all information about the RSE might be lost!"""
    disable_rse(Arguments({"rse": rse_name}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@rse.command("update")
@click.argument("rse-name")
@click.option('--key', help='Setting key', required=True)
@click.option('--value', help='Setting value', required=True)
@click.pass_context
def update(ctx, rse_name, key, value):
    """
    Update an RSE's setting.

    \b
    Example:
        $ rucio rse update my-rse --option availability_write True
    """
    args = Arguments({"rse": rse_name, "param": key, "value": value})
    update_rse(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@rse.group()
@click.help_option("-h", "--help")
def distance():
    """Manage the relative distance between RSEs for transfer prioritization calculations"""


@distance.command("show")
@click.argument("source-rse")
@click.argument("destination-rse")
@click.pass_context
def distance_show(ctx, source_rse, destination_rse):
    """Display distance information from SOURCE-RSE to DESTINATION-RSE"""
    get_distance_rses(Arguments({"source": source_rse, "destination": destination_rse}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@distance.command("add")
@click.argument("source-rse")
@click.argument("destination-rse")
@click.option("--distance", default=1, type=int, help="Relative distance between RSEs")
@click.pass_context
def distance_add(ctx, source_rse, destination_rse, distance):
    """Create a new link from SOURCE-RSE to DESTINATION-RSE with a distance"""
    args = Arguments({"source": source_rse, "destination": destination_rse, "distance": distance})
    add_distance_rses(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@distance.command("remove")
@click.argument("source-rse")
@click.argument("destination-rse")
@click.pass_context
def distance_remove(ctx, source_rse, destination_rse):
    """Un-link SOURCE-RSE from DESTINATION-RSE by removing the distance between them"""
    args = Arguments({"source": source_rse, "destination": destination_rse})
    delete_distance_rses(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@distance.command("update")
@click.argument("source-rse")
@click.argument("destination-rse")
@click.option("--distance", type=int, help="Relative distance between RSEs", required=True)
@click.pass_context
def distance_update(ctx, source_rse, destination_rse, distance):
    """Update the existing distance or ranking from SOURCE-RSE to DESTINATION-RSE"""
    args = Arguments({"source": source_rse, "destination": destination_rse, "distance": distance})
    update_distance_rses(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@rse.group()
def attribute():
    """Interact with RSE Attributes"""


@attribute.command("list")
@click.argument("rse-name")
@click.pass_context
def attr_list_(ctx, rse_name):
    """List all attributes of a given RSE"""
    get_attribute_rse(Arguments({"rse": rse_name}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@attribute.command("add")
@click.argument("rse-name")
@click.option('--key', help='Attribute key', required=True)
@click.option('--value', help='Attribute value', required=True)
@click.pass_context
def attribute_add_(ctx, rse_name, key, value):
    """Add a new attribute for an RSE

    \b
    Example:
        $ rucio rse attribute add my-rse --key My-Attribute  --value True
    """
    args = Arguments({"rse": rse_name, "key": key, "value": value})
    set_attribute_rse(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


# TODO Update attribute - only overwrites existing attributes

@attribute.command("remove")
@click.argument("rse-name")
@click.option("--attribute", help="Attribute to remove", required=True)
@click.pass_context
def attribute_remove(ctx, rse_name, attribute):
    """Remove an existing attribute from an RSE"""
    args = Arguments({"rse": rse_name, "key": attribute, "value": None})
    delete_attribute_rse(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@rse.group()
def limit():
    """Manage storage size limits"""


@limit.command("add")
@click.argument("rse-name")
@click.option("--limit", type=(str, int), required=True, help="Name of limit and value in bytes")
@click.pass_context
def limit_add(ctx, rse_name, limit):
    """Add a usage limit to an RSE

    \b
    Example, add a limit of 1KB to XRD1 named "MinFreeSpace":
        $ rucio rse limit add XRD1 --limit MinFreeSpace 10000
    """
    args = Arguments({"rse": rse_name, "name": limit[0], "value": limit[1]})
    set_limit_rse(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@limit.command("remove")
@click.argument("rse-name")
@click.option("--limit", required=True, help="Name of limit to remove")
@click.pass_context
def limit_remove(ctx, rse_name, limit):
    """Remove an existing RSE limit"""
    args = Arguments({"rse": rse_name, "name": limit})
    delete_limit_rse(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@rse.group()
@click.help_option("-h", "--help")
def protocol():
    """Manage RSE transfer protocols"""


# TODO Better loader for json types
@protocol.command("add")
@click.argument("rse-name")
@click.option("--host", "--host-name", help="Endpoint hostname", required=True)
@click.option("--scheme", help="Endpoint URL scheme", required=True)
@click.option("--prefix", help="Endpoint URL path prefix", required=True)
@click.option("--space-token", help="Space token name (SRM-only)")
@click.option("--web-service-path", help="Web service URL (SRM-only)")
@click.option("--port", type=int, help="URL port")
@click.option("--impl", default="rucio.rse.protocols.gfal.Default", help="Transfer protocol implementation to use")
@click.option("--domain-json", type=json.loads, help="JSON describing the WAN / LAN setup")
@click.option("--extended-attributes-json", type=json.loads, help="JSON describing any extended attributes")
@click.pass_context
def protocol_add(ctx, rse_name, host, scheme, prefix, space_token, web_service_path, port, impl, domain_json, extended_attributes_json):
    """
    Add a new protocol for an RSE used for transferring files

    \b
    Example, adding a default protocol hosted at jdoes.test.org to the RSE JDOE_DATADISK
        $ rucio rse protocol add JDOE_DATADISK --host-name jdoes.test.org --scheme gsiftp --prefix '/atlasdatadisk/rucio/' --port 8443'

    """
    args = Arguments(
        {"rse": rse_name, "hostname": host, "ext_attr_json": extended_attributes_json, "scheme": scheme, "prefix": prefix, "space_token": space_token, "web_service_path": web_service_path, "port": port, "impl": impl, "domain_json": domain_json}
    )
    add_protocol_rse(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@protocol.command("remove")
@click.argument("rse-name")
@click.option("--scheme", help="Endpoint URL scheme", required=True)
@click.option("--host-name", help="Endpoint hostname")
@click.option("--port", type=int, help="URL port")
@click.pass_context
def protocol_remove(ctx, rse_name, host_name, scheme, port):
    """Remove an existing protocol from an RSE"""
    args = Arguments({"rse": rse_name, "scheme": scheme, "hostname": host_name, "port": port})
    del_protocol_rse(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@rse.group()
@click.help_option("-h", "--help")
def qos():
    """Interact with the QoS model"""


@qos.command("add")
@click.argument("rse-name", nargs=1)
@click.option("--policy", required=True)
@click.pass_context
def qos_add(ctx, rse_name, policy):
    "Add a new QoS policy"
    add_qos_policy(Arguments({"rse": rse_name, "qos_policy": policy}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@qos.command("remove")
@click.argument("rse-name", nargs=1)
@click.option("--policy", required=True)
@click.pass_context
def qos_remove(ctx, rse_name, policy):
    "Remove an existing QoS policy"
    delete_qos_policy(Arguments({"rse": rse_name, "qos_policy": policy}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@qos.command("list")
@click.argument("rse-name", nargs=1)
@click.pass_context
def qos_list(ctx, rse_name):
    "List the RSE's QoS policies"
    list_qos_policies(Arguments({"rse": rse_name}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import subprocess
import sys
from typing import Union

clients_requirements_table = {
    'install_requires': [
        'requests',
        'urllib3',
        'dogpile-cache<=1.2.2',
        'packaging',
        'tabulate',
        'jsonschema',
        'dataclasses',
        'rich',
        'typing_extensions'
    ],
    'extras_require': {
        'ssh': ['paramiko'],
        'kerberos': [
            'kerberos',
            'pykerberos',
            'requests-kerberos',
        ],
        'swift': ['python-swiftclient'],
        'argcomplete': ['argcomplete'],
        'sftp': ['paramiko'],
        'dumper': [
            'python-magic',
        ],
    }
}

dev_requirements = [
    'pytest',
    'pytest-xdist',
    'pytest-cov',
    'pyflakes',
    'flake8',
    'pylint',
    'isort',
    'xmltodict',
    'pytz',
    'pycodestyle',
    'pydoc-markdown',
    'docspec_python',
    'sh',
    'PyYAML',
]

server_requirements_table = {
    'install_requires': clients_requirements_table['install_requires'] + [
        'argcomplete',
        'boto',
        'python-magic',
        'paramiko',
        'boto3',
        'sqlalchemy',
        'alembic',
        'pymemcache',
        'python-dateutil',
        'stomp-py',
        'statsd',
        'geoip2',
        'google-auth',
        'redis',
        'flask',
        'oic',
        'prometheus_client',
    ],
    'extras_require': {
        'oracle': ['cx_oracle'],
        'mongo': ['pymongo'],
        'elastic': ['elasticsearch'],
        'postgresql': ['psycopg[binary,pool]'],
        'mysql': ['PyMySQL'],
        'kerberos': [
            'kerberos',
            'pykerberos',
            'requests-kerberos',
        ],
        'globus': [
            'PyYAML',
            'globus-sdk',
        ],
        'saml': ['python3-saml'],
        'dev': dev_requirements
    }
}


def run_shell_command(cmd: str) -> Union[str, bytearray, memoryview]:
    """
    Run a shell command in path and return output"

    :param cmd: the shell command.
    :return: Output of the shell command.
    """
    output = subprocess.Popen(["/bin/sh", "-c", cmd], stdout=subprocess.PIPE)
    stdout = output.communicate()[0].strip()
    if isinstance(stdout, bytes):
        stdout = stdout.decode(errors='replace')
    return stdout


def get_rucio_version() -> str:
    python_executable = "'" + sys.executable + "'"
    ver = run_shell_command(
        "PYTHONPATH=lib " + python_executable + " -c "
        '"from rucio import version; print(version.version_string())"'
    )
    if not ver:
        raise RuntimeError("Could not fetch Rucio version")
    return str(ver)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic decommissioning profiles."""
import logging
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any

from sqlalchemy.exc import NoResultFound

from rucio.common.constants import RseAttr
from rucio.common.exception import Duplicate, ReplicaNotFound, RequestNotFound, RucioException, RuleNotFound, RuleReplaceFailed, UnsupportedOperation
from rucio.core.lock import get_replica_locks, get_replica_locks_for_rule_id
from rucio.core.replica import list_replicas_per_rse, set_tombstone, update_replica_state
from rucio.core.request import get_request_by_did
from rucio.core.rse import add_rse_attribute, update_rse
from rucio.core.rule import list_rules_for_rse_decommissioning, move_rule, update_rule
from rucio.db.sqla.constants import ReplicaState

from .types import DecommissioningProfile, HandlerOutcome

if TYPE_CHECKING:
    from collections.abc import Iterable

    from rucio.common.types import LoggerFunction


def generic_delete(rse: dict[str, Any], config: dict[str, Any]) -> DecommissioningProfile:
    """Return a profile for simply deleting rules.

    The "generic delete" profile lists out all rules that are locking replicas at the given RSE,
    and deletes them if either one of the following is true:

    - The RSE expression of the rule is trivial (the RSE name itself).
    - There are no replicas locked by the rule that reside on another RSE.

    :param rse: RSE to decommission.
    :param config: Decommissioning configuration dictionary.
    :returns: A decommissioning profile.
    """
    return DecommissioningProfile(
        rse=rse,
        initializer=_generic_initialize,
        discoverer=_generic_discover,
        handlers=[
            (_is_locked, _call_for_attention),
            (_is_being_deleted, _count_as_processed),
            (_has_trivial_rse_expression, _delete_rule),
            (_has_all_replicas_on_rse, _delete_rule),
        ],
        finalizer=_generic_finalize
    )


def generic_move(rse: dict[str, Any], config: dict[str, Any]) -> DecommissioningProfile:
    """Return a profile for moving rules that satisfy conditions to a specific destination.

    The "generic move" profile lists out all rules that are locking replicas at the given RSE,
    and moves them to the specified destination if either one of the following is true:

    - The RSE expression of the rule is trivial (the RSE name itself).
    - There are no replicas locked by the rule that reside on another RSE.

    :param rse: RSE to decommission.
    :param config: Decommissioning configuration dictionary.
    :returns: A decommissioning profile dictionary.
    """
    try:
        destination = config['destination']
    except KeyError as exc:
        raise RucioException('Configuration parameter "destination" not set') from exc

    move_to_destination = RuleMover(destination)
    return DecommissioningProfile(
        rse=rse,
        initializer=_generic_initialize,
        discoverer=_generic_discover,
        handlers=[
            (_is_locked, _call_for_attention),
            (_is_being_deleted, _count_as_processed),
            (_has_child_rule_id, _count_as_processed),
            (_has_trivial_rse_expression, move_to_destination),
            (_has_all_replicas_on_rse, move_to_destination),
        ],
        finalizer=_generic_finalize
    )


def _generic_initialize(
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log,
) -> None:
    """Initializer function that sets the RSE availability flags and the decommissioning status.

    When an RSE is initialized for decommissioning, the availability flags are set as follows:

    - `availability_read=True`
    - `availability_write=False`
    - `availability_delete=True`

    and the following attribute is set:

    - `greedyDeletion=True`

    :param rse: RSE table entry as a dictionary.
    :param logger: Logging function.
    """
    logger(logging.INFO,
           '(%s) Setting RSE availability flags of to read !write delete, greedyDeletion=true.',
           rse['rse'])

    parameters = {
        'availability_read': True,
        'availability_write': False,
        'availability_delete': True
    }
    update_rse(rse['id'], parameters)

    try:
        add_rse_attribute(rse['id'], RseAttr.GREEDYDELETION, True)
    except Duplicate:
        pass


def _generic_discover(
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> 'Iterable[dict[str, Any]]':
    """Discoverer function that calls the listing function from core.rule.

    :param rse: RSE table entry as a dictionary.
    :param logger: Logging function.
    :returns: An iterable of rule dictionaries.
    """
    return list_rules_for_rse_decommissioning(rse['id'])


def _generic_finalize(
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> bool:
    """Check for remaining replicas at the RSE and resolve where possible.

    If for some reason there are still a large number of replicas remaining at the RSE, this
    function cuts off automatic processing attempts at a hard-coded 100.

    :param rse: RSE table entry as a dictionary.
    :param logger: Logging function.
    :returns: A boolean flag indicating the completion status.
    """
    num_remaining_replicas = _process_replicas_with_no_locks(rse,
                                                             list_replicas_per_rse(rse['id']),
                                                             limit=100,
                                                             logger=logger)

    if num_remaining_replicas == 0:
        # The RSE is now really completely cleared.
        logger(logging.INFO,
               '(%s) Completed deleting the RSE contents. Updating the decommissioning status'
               ' to "done".',
               rse['rse'])
        return True

    logger(logging.WARNING,
           '(%s) %d replicas remain on the RSE even though there are no more rules creating'
           ' locks on them.',
           rse['rse'], num_remaining_replicas)

    # We now wait for reaper to pick the updated replicas (if there are any). Hopefully all
    # replicas will be deleted at the next cycle.
    return False


def _process_replicas_with_no_locks(
    rse: dict[str, Any],
    replicas: 'Iterable[dict[str, Any]]',
    limit: int = 0,
    *,
    logger: "LoggerFunction" = logging.log,
) -> int:
    """Process replicas that remain at the RSE after the decommissioning.

    Replicas that remain at the RSE at this point should be unlocked. While this is a situation
    that should not happen, it can due to various errors. This function checks each such replica
    and attempt to resolve the invalid state for known cases.

    :param rse: RSE table entry as a dictionary.
    :param replicas: Sequence or generator of replica objects.
    :param logger: Logging function.
    :returns: Number of processed replicas.
    """
    num_processed = 0
    num_updated = 0
    num_waiting_reap = 0

    # Check if each replica satisfies the condition in list_and_mark_unlocked_replicas() (in
    # core.replica).
    # For each of the conditions, in case of failure, handle cases that can be handled.
    for replica in replicas:
        num_processed += 1
        if limit > 0 and num_processed == limit:
            break

        updated = False

        # If the replica fails any of the following three conditions, reaper won't pick it up.
        reap_pending = True

        # 1. Condition on the replica state & update time
        in_stable_state = replica['state'] in [ReplicaState.AVAILABLE,
                                               ReplicaState.UNAVAILABLE,
                                               ReplicaState.BAD]

        ten_minutes_ago = datetime.utcnow() - timedelta(seconds=600)
        deleting_stuck = (replica['state'] == ReplicaState.BEING_DELETED
                          and replica['updated_at'] < ten_minutes_ago)

        if not (in_stable_state or deleting_stuck):
            # 1.1 If COPYING without a valid request
            # -> Update the state to UNAVAILABLE
            if replica['state'] == ReplicaState.COPYING:
                # Should we allow requests in certain states? Depends on the details of the
                # request state machine.
                try:
                    get_request_by_did(replica['scope'], replica['name'], rse['id'])
                except RequestNotFound:
                    logger(logging.INFO,
                           '(%s) Replica %s:%s is in COPYING state without a valid request.'
                           ' Changing state to UNAVAILABLE.',
                           rse['rse'], replica['scope'], replica['name'])

                    try:
                        update_replica_state(rse['id'], replica['scope'], replica['name'],
                                             ReplicaState.UNAVAILABLE)
                    except ReplicaNotFound:
                        logger(logging.WARNING,
                               '(%s) Replica %s:%s disappeared during cleanup',
                               rse['rse'], replica['scope'], replica['name'])
                    except UnsupportedOperation as error:
                        logger(logging.ERROR, '(%s) %s', rse['rse'], str(error))
                    else:
                        updated = True

            reap_pending = False

        # 2. Condition on the lock count
        if replica['lock_cnt'] != 0:
            # 2.1 No actual lock -> Reset the lock count
            try:
                locks = get_replica_locks(replica['scope'], replica['name'],
                                          restrict_rses=[rse['id']])
            except NoResultFound:
                logger(logging.WARNING,
                       '(%s) Replica %s:%s disappeared during cleanup',
                       rse['rse'], replica['scope'], replica['name'])
            else:
                if len(locks) == 0:
                    logger(logging.WARNING,
                           '(%s) Replica %s:%s has lock count %s but zero actual locks. Please'
                           ' fix the counts.',
                           rse['rse'], replica['scope'], replica['name'], replica['lock_cnt'])

            reap_pending = False

        # 3. Tombstone missing or in the future -> RIP now
        if replica['tombstone'] is None or replica['tombstone'] >= datetime.utcnow():
            logger(logging.INFO, '(%s) Marking tombstone of replica %s:%s as UTCNOW.',
                   rse['rse'], replica['scope'], replica['name'])

            try:
                set_tombstone(rse['id'], replica['scope'], replica['name'],
                              tombstone=datetime.utcnow())
            except ReplicaNotFound as error:
                logger(logging.WARNING, '(%s) %s', rse['rse'], str(error))
            else:
                updated = True

            reap_pending = False

        if updated:
            num_updated += 1

        if reap_pending:
            num_waiting_reap += 1

    logger(logging.INFO, '(%s) %s replicas have been updated. %s replicas are pending deletion.'
           ' >=%s replicas remain unhandled.',
           rse['rse'], num_updated, num_waiting_reap,
           num_processed - num_updated - num_waiting_reap)

    return num_processed


# Condition functions

def _is_locked(
    rule: dict[str, Any],
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> bool:
    logger(logging.INFO,
           '(%s) Rule %s for %s:%s is locked',
           rse['rse'], rule['id'], rule['scope'], rule['name'])
    return rule['locked']


def _is_being_deleted(
    rule: dict[str, Any],
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> bool:
    """Check if the rule is already set to be deleted."""
    if rule['expires_at'] is not None and rule['expires_at'] < datetime.utcnow():
        logger(logging.DEBUG,
               '(%s) Rule %s for %s:%s is bound for deletion',
               rse['rse'], rule['id'], rule['scope'], rule['name'])
        return True
    return False


def _has_trivial_rse_expression(
    rule: dict[str, Any],
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> bool:
    """Check for a trivial rse_expression."""
    return rule['rse_expression'] == rse['rse']


def _has_all_replicas_on_rse(
    rule: dict[str, Any],
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> bool:
    """Check if the all the replicas are on the RSE."""
    # Check the list of RSEs that the replicas locked by this rule reside on.
    try:
        locks = get_replica_locks_for_rule_id(rule['id'])
    except NoResultFound:
        # No replica is locked to begin with -> treat as having the last replica on this RSE
        return True

    replica_rse_ids = set(lock['rse_id'] for lock in locks)

    return len(replica_rse_ids) == 1 and list(replica_rse_ids)[0] == rse['id']


def _has_child_rule_id(
    rule: dict[str, Any],
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> bool:
    """Check for non-empty child_rule_id."""
    if rule['child_rule_id']:
        logger(logging.DEBUG, '(%s) Rule %s for %s:%s is being moved',
               rse['rse'], rule['id'], rule['scope'], rule['name'])
        return True
    return False


# Action functions

def _call_for_attention(
    rule: dict[str, Any],
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> HandlerOutcome:
    return HandlerOutcome.NEED_ATTENTION


def _count_as_processed(
    rule: dict[str, Any],
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> HandlerOutcome:
    """Do nothing but increment the limit checker."""
    return HandlerOutcome.UNTOUCHED


def _delete_rule(
    rule: dict[str, Any],
    rse: dict[str, Any],
    *,
    logger: "LoggerFunction" = logging.log
) -> HandlerOutcome:
    """Delete the rule."""
    logger(logging.DEBUG, '(%s) Deleting rule %s for %s:%s',
           rse['rse'], rule['id'], rule['scope'], rule['name'])

    try:
        update_rule(rule['id'], {'lifetime': 0})
    except RuleNotFound:
        logger(logging.WARNING, '(%s) Rule %s for %s:%s disappeared before deleting',
               rse['rse'], rule['id'], rule['scope'], rule['name'])
        return HandlerOutcome.UNTOUCHED

    return HandlerOutcome.REMOVED


class RuleMover:
    """Callable object with a destination attribute."""
    def __init__(self, destination: str) -> None:
        self.destination = destination

    def __call__(
        self,
        rule: dict[str, Any],
        rse: dict[str, Any],
        *,
        logger: "LoggerFunction" = logging.log
    ) -> HandlerOutcome:
        """Move the rule."""
        # Move the rule.
        logger(logging.DEBUG, '(%s) Moving rule %s for %s:%s to %s',
               rse['rse'], rule['id'], rule['scope'], rule['name'],
               self.destination)

        try:
            move_rule(rule['id'], self.destination,
                      override={'weight': None, 'source_replica_expression': None})
        except RuleNotFound:
            logger(logging.WARNING, '(%s) Rule %s for %s:%s disappeared before moving',
                   rse['rse'], rule['id'], rule['scope'], rule['name'])
            return HandlerOutcome.UNTOUCHED
        except RuleReplaceFailed:
            logger(logging.ERROR,
                   '(%s) Failed to move rule %s for %s:%s to %s',
                   rse['rse'], rule['id'], rule['scope'], rule['name'],
                   self.destination)
            return HandlerOutcome.NEED_ATTENTION
        return HandlerOutcome.REMOVED



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import os

import pytest

from rucio.common.bittorrent import bittorrent_v2_merkle_sha256
from rucio.common.exception import InvalidType
from rucio.common.logging import formatted_logger
from rucio.common.utils import Availability, parse_did_filter_from_string, retrying


class TestUtils:
    """UTILS (COMMON): test utilisty functions"""

    def test_parse_did_filter_string(self):
        """(COMMON/UTILS): test parsing of did filter string"""
        test_cases = [{
            'input': 'type=all,length=3,length>4,length>=6,length<=7,  test=b, created_after=1900-01-01T00:00:00.000Z',
            'expected_filter': {'length': 3, 'length.gt': 4, 'length.gte': 6, 'length.lte': 7, 'test': 'b', 'created_after': datetime.datetime.strptime('1900-01-01T00:00:00.000Z', '%Y-%m-%dT%H:%M:%S.%fZ')},
            'expected_type': 'all'
        }, {
            'input': 'type=FILE',
            'expected_filter': {},
            'expected_type': 'file'
        }, {
            'input': '',
            'expected_filter': {},
            'expected_type': 'collection'
        }]

        for test_case in test_cases:
            filters, type_ = parse_did_filter_from_string(test_case['input'])
            assert test_case['expected_filter'] == filters
            assert test_case['expected_type'] == type_

        with pytest.raises(InvalidType):
            input_ = 'type=g'
            parse_did_filter_from_string(input_)

    def test_availability_data_class(self):
        availability = Availability()

        assert availability.read is None
        assert availability.write is None
        assert availability.delete is None

        availability = Availability(True, False, True)

        assert availability.read
        assert not availability.write
        assert availability.delete

    def test_availability_tuple_unpacking(self):
        read, write, delete = Availability(True, False, True)

        assert read
        assert not write
        assert delete

    def test_availability_hash(self):
        hash(Availability(True, True, True))

    def test_availability_with_none(self):
        assert Availability(write=False).integer == 5

    def test_availability_from_integer_none(self):
        assert Availability.from_integer(None) == Availability(None, None, None)


@pytest.mark.parametrize(
    "integer,tuple_values",
    [
        #   (read,  write, delete)
        (7, (None, None, None)),
        (6, (True, None, False)),
        (5, (None, False, None)),
    ],
)
def test_availability_convert_with_none(integer, tuple_values):
    """
    This tests the conversion to an integer with missing values. Missing values
    should be interpreted as `True`, since this is the default value.
    """
    assert integer == Availability(*tuple_values).integer


@pytest.mark.parametrize(
    "before,after",
    [
        #   (read,  write, delete)
        (7, (True, True, True)),
        (6, (True, True, False)),
        (5, (True, False, True)),
        (4, (True, False, False)),
        (3, (False, True, True)),
        (2, (False, True, False)),
        (1, (False, False, True)),
        (0, (False, False, False)),
    ],
)
def test_availability_translation(before, after):
    assert Availability.from_integer(before) == Availability(*after)
    assert tuple(Availability.from_integer(before)) == after

    assert Availability.from_integer(before).integer == before
    assert Availability(*after).integer == before


def test_formatted_logger():
    result = None

    def log_func(level, msg, *args, **kwargs):
        nonlocal result
        result = (level, msg)

    new_log_func = formatted_logger(log_func, "a %s c")

    new_log_func(logging.INFO, "b")
    assert result == (logging.INFO, "a b c")


def test_retrying():

    attempts = []
    start_time = datetime.datetime.now()

    @retrying(retry_on_exception=lambda _: True, wait_fixed=550, stop_max_attempt_number=3)
    def always_retry():
        attempts.append(True)
        raise ValueError()

    with pytest.raises(ValueError):
        always_retry()
    assert len(attempts) == 3
    assert datetime.datetime.now() - start_time > datetime.timedelta(seconds=1)

    attempts = []

    @retrying(retry_on_exception=lambda e: isinstance(e, AttributeError), wait_fixed=1, stop_max_attempt_number=3)
    def retry_on_attribute_error():
        attempts.append(True)
        raise ValueError()

    with pytest.raises(ValueError):
        retry_on_attribute_error()
    assert len(attempts) == 1


def test_bittorrent_sa256_merkle(file_factory):

    def _sha256_merkle_via_libtorrent(file, piece_size=0):
        import libtorrent as lt
        file = str(file)
        fs = lt.file_storage()
        lt.add_files(fs, file)
        t = lt.create_torrent(fs, flags=lt.create_torrent.v2_only, piece_size=piece_size)
        lt.set_piece_hashes(t, os.path.dirname(file))

        torrent = t.generate()
        pieces_root = next(iter(next(iter(torrent[b'info'][b'file tree'].values())).values()))[b'pieces root']
        pieces_layers = torrent.get(b'piece layers', {}).get(pieces_root, b'')
        piece_size = t.piece_length()

        return pieces_root, pieces_layers, piece_size

    for size in (1, 333, 1024, 16384, 16390, 32768, 32769, 49152, 65530, 65536, 81920, 2**20 - 2**17, 2**20, 2**20 + 2):
        file = file_factory.file_generator(size=size)
        root, layers, piece_size = bittorrent_v2_merkle_sha256(file)
        assert (root, layers, piece_size) == _sha256_merkle_via_libtorrent(file, piece_size=piece_size)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Union

from sqlalchemy.exc import DatabaseError
from sqlalchemy.sql.expression import and_, or_, select, true, update

import rucio.core.did
import rucio.core.rule
from rucio.common.constants import RseAttr
from rucio.common.exception import DataIdentifierNotFound
from rucio.common.types import InternalScope, LoggerFunction
from rucio.core.lifetime_exception import define_eol
from rucio.core.rse import get_rse_attribute, get_rse_name
from rucio.db.sqla import filter_thread_work, models
from rucio.db.sqla.constants import DIDType, LockState, RuleGrouping, RuleNotification, RuleState
from rucio.db.sqla.session import read_session, stream_session, transactional_session

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from sqlalchemy.orm import Session


@stream_session
def get_dataset_locks(scope: InternalScope, name: str, *, session: "Session") -> "Iterator[dict[str, Any]]":
    """
    Get the dataset locks of a dataset

    :param scope:          Scope of the dataset.
    :param name:           Name of the dataset.
    :param session:        The db session.
    :return:               List of dicts {'rse_id': ..., 'state': ...}
    """

    stmt = select(
        models.DatasetLock.rse_id,
        models.DatasetLock.scope,
        models.DatasetLock.name,
        models.DatasetLock.rule_id,
        models.DatasetLock.account,
        models.DatasetLock.state,
        models.DatasetLock.length,
        models.DatasetLock.bytes,
        models.DatasetLock.accessed_at
    ).where(
        and_(models.DatasetLock.scope == scope,
             models.DatasetLock.name == name)
    )

    for rse_id, scope, name, rule_id, account, state, length, bytes_, accessed_at in session.execute(stmt).yield_per(500):
        yield {'rse_id': rse_id,
               'rse': get_rse_name(rse_id, session=session),
               'scope': scope,
               'name': name,
               'rule_id': rule_id,
               'account': account,
               'state': state,
               'length': length,
               'bytes': bytes_,
               'accessed_at': accessed_at}


@stream_session
def get_dataset_locks_bulk(dids: "Iterable[dict[str, Any]]", *, session: "Session") -> "Iterator[dict[str, Any]]":
    """
    Get the dataset locks of a list of datasets or containers, recursively

    :param dids:           List of dictionaries {"scope":scope(type:InternalScope), "name":name,
                           "type":did type(DIDType.DATASET or DIDType.CONTAINER)}, "type" is optional
    :param session:        The db session to use.
    :return:               Generator of lock_info dicts, may contain duplicates
    """

    for did in dids:
        scope = did["scope"]
        if not isinstance(scope, InternalScope):
            raise ValueError("Scope must be passed as an InternalScope object.")
        name = did["name"]
        did_type = did.get("type")
        if not did_type:
            try:
                did_info = rucio.core.did.get_did(scope, name, session=session)
            except DataIdentifierNotFound:
                continue
            did_type = did_info["type"]

        if did_type == DIDType.DATASET:
            for lock_dict in get_dataset_locks(scope, name, session=session):
                yield lock_dict
        elif did_type == DIDType.CONTAINER:
            for dataset_info in rucio.core.did.list_child_datasets(scope, name, session=session):
                dataset_scope, dataset_name = dataset_info["scope"], dataset_info["name"]
                for lock_dict in get_dataset_locks(dataset_scope, dataset_name, session=session):
                    yield lock_dict
        else:
            raise ValueError("Can only get locks for datasets and containers.")


@stream_session
def get_dataset_locks_by_rse_id(rse_id: str, *, session: "Session") -> "Iterator[dict[str, Any]]":
    """
    Get the dataset locks of an RSE.

    :param rse_id:         RSE id to get the locks from.
    :param session:        The db session.
    :return:               List of dicts {'rse_id': ..., 'state': ...}
    """
    stmt = select(
        models.DatasetLock.rse_id,
        models.DatasetLock.scope,
        models.DatasetLock.name,
        models.DatasetLock.rule_id,
        models.DatasetLock.account,
        models.DatasetLock.state,
        models.DatasetLock.length,
        models.DatasetLock.bytes,
        models.DatasetLock.accessed_at
    ).with_hint(
        models.DatasetLock,
        'INDEX(DATASET_LOCKS DATASET_LOCKS_RSE_ID_IDX)',
        'oracle'
    ).where(
        models.DatasetLock.rse_id == rse_id
    )

    for rse_id, scope, name, rule_id, account, state, length, bytes_, accessed_at in session.execute(stmt).yield_per(500):
        yield {'rse_id': rse_id,
               'rse': get_rse_name(rse_id, session=session),
               'scope': scope,
               'name': name,
               'rule_id': rule_id,
               'account': account,
               'state': state,
               'length': length,
               'bytes': bytes_,
               'accessed_at': accessed_at}


@read_session
def get_replica_locks(scope: InternalScope, name: str, nowait: bool = False, restrict_rses: Optional["Iterable[str]"] = None, *, session: "Session") -> list[models.ReplicaLock]:
    """
    Get the active replica locks for a file

    :param scope:          Scope of the did.
    :param name:           Name of the did.
    :param nowait:         Nowait parameter for the FOR UPDATE statement.
    :param restrict_rses:  Possible RSE_ids to filter on.
    :param session:        The db session.
    :return:               List of dicts {'rse': ..., 'state': ...}
    :raises:               NoResultFound
    """

    stmt = select(
        models.ReplicaLock
    ).where(
        and_(models.ReplicaLock.scope == scope,
             models.ReplicaLock.name == name)
    )
    if restrict_rses is not None:
        rse_clause = []
        for rse_id in restrict_rses:
            rse_clause.append(models.ReplicaLock.rse_id == rse_id)
        if rse_clause:
            stmt = stmt.where(
                or_(*rse_clause)
            )

    stmt = stmt.with_for_update(
        nowait=nowait
    )
    return list(session.execute(stmt).scalars().all())


@read_session
def get_replica_locks_for_rule_id(rule_id: str, *, session: "Session") -> list[dict[str, Any]]:
    """
    Get the active replica locks for a rule_id.

    :param rule_id:        Filter on rule_id.
    :param session:        The db session.
    :return:               List of dicts {'scope':, 'name':, 'rse': ..., 'state': ...}
    :raises:               NoResultFound
    """

    locks = []

    stmt = select(
        models.ReplicaLock
    ).where(
        models.ReplicaLock.rule_id == rule_id
    )

    for row in session.execute(stmt).scalars().all():
        locks.append({'scope': row.scope,
                      'name': row.name,
                      'rse_id': row.rse_id,
                      'rse': get_rse_name(rse_id=row.rse_id, session=session),
                      'state': row.state,
                      'rule_id': row.rule_id})

    return locks


@read_session
def get_replica_locks_for_rule_id_per_rse(rule_id: str, *, session: "Session") -> list[dict[str, str]]:
    """
    Get the active replica locks for a rule_id per rse.

    :param rule_id:        Filter on rule_id.
    :param session:        The db session.
    :return:               List of dicts {'rse_id':, 'rse':}
    :raises:               NoResultFound
    """

    locks = []

    stmt = select(
        models.ReplicaLock.rse_id
    ).where(
        models.ReplicaLock.rule_id == rule_id
    ).group_by(
        models.ReplicaLock.rse_id
    )

    for rse_id in session.execute(stmt).scalars().all():
        locks.append({'rse_id': rse_id,
                      'rse': get_rse_name(rse_id=rse_id, session=session)})

    return locks


@read_session
def get_files_and_replica_locks_of_dataset(scope: InternalScope, name: str, nowait: bool = False, restrict_rses: Optional["Iterable[str]"] = None, only_stuck: bool = False,
                                           total_threads: Optional[int] = None, thread_id: Optional[int] = None,
                                           *, session: "Session") -> dict[tuple[InternalScope, str], Union[models.ReplicaLock, list[models.ReplicaLock]]]:
    """
    Get all the files of a dataset and, if existing, all locks of the file.

    :param scope:          Scope of the dataset
    :param name:           Name of the dataset
    :param nowait:         Nowait parameter for the FOR UPDATE statement
    :param restrict_rses:  Possible RSE_ids to filter on.
    :param only_stuck:     If true, only get STUCK locks.
    :param total_threads:  Total threads
    :param thread_id:      This thread
    :param session:        The db session.
    :return:               Dictionary with keys: (scope, name)
                           and as value: [LockObject]
    :raises:               NoResultFound
    """
    locks = {}

    rse_clause = [true()]
    if restrict_rses is not None:
        rse_clause = [models.ReplicaLock.rse_id == rse_id for rse_id in restrict_rses]

    base_stmt = select(
        models.DataIdentifierAssociation.child_scope,
        models.DataIdentifierAssociation.child_name
    ).where(
        and_(models.DataIdentifierAssociation.scope == scope,
             models.DataIdentifierAssociation.name == name)
    )
    stmt = base_stmt.add_columns(
        models.ReplicaLock
    ).with_for_update(
        nowait=nowait,
        of=models.ReplicaLock.state
    )

    if session.bind.dialect.name == 'postgresql':
        if total_threads and total_threads > 1:
            base_stmt = filter_thread_work(session=session, query=base_stmt, total_threads=total_threads,
                                           thread_id=thread_id, hash_variable='child_name')

        for child_scope, child_name in session.execute(base_stmt).yield_per(1000):
            locks[(child_scope, child_name)] = []

        stmt = stmt.where(
            and_(models.DataIdentifierAssociation.child_scope == models.ReplicaLock.scope,
                 models.DataIdentifierAssociation.child_name == models.ReplicaLock.name,
                 or_(*rse_clause))
        )
    else:
        stmt = stmt.with_hint(
            models.DataIdentifierAssociation,
            'INDEX_RS_ASC(CONTENTS CONTENTS_PK) NO_INDEX_FFS(CONTENTS CONTENTS_PK)',
            'oracle'
        ).outerjoin(
            models.ReplicaLock,
            and_(models.DataIdentifierAssociation.child_scope == models.ReplicaLock.scope,
                 models.DataIdentifierAssociation.child_name == models.ReplicaLock.name,
                 or_(*rse_clause))
        )

    if only_stuck:
        stmt = stmt.where(
            models.ReplicaLock.state == LockState.STUCK
        )

    if total_threads and total_threads > 1:
        stmt = filter_thread_work(session=session, query=stmt, total_threads=total_threads,
                                  thread_id=thread_id, hash_variable='child_name')

    for child_scope, child_name, lock in session.execute(stmt).all():
        if (child_scope, child_name) not in locks:
            if lock is None:
                locks[(child_scope, child_name)] = []
            else:
                locks[(child_scope, child_name)] = [lock]
        else:
            locks[(child_scope, child_name)].append(lock)

    return locks


@transactional_session
def successful_transfer(scope: InternalScope, name: str, rse_id: str, nowait: bool, *, session: "Session", logger: LoggerFunction = logging.log) -> None:
    """
    Update the state of all replica locks because of an successful transfer

    :param scope:    Scope of the did
    :param name:     Name of the did
    :param rse_id:   RSE id
    :param nowait:   Nowait parameter for the for_update queries.
    :param session:  DB Session.
    """

    stmt = select(
        models.ReplicaLock
    ).where(
        and_(models.ReplicaLock.scope == scope,
             models.ReplicaLock.name == name,
             models.ReplicaLock.rse_id == rse_id)
    ).with_for_update(
        nowait=nowait
    )
    for lock in session.execute(stmt).scalars().all():
        if lock.state == LockState.OK:
            continue
        logger(logging.DEBUG, 'Marking lock %s:%s for rule %s on rse %s as OK' % (lock.scope, lock.name, str(lock.rule_id), get_rse_name(rse_id=lock.rse_id, session=session)))
        # Update the rule counters
        stmt = select(
            models.ReplicationRule
        ).where(
            models.ReplicationRule.id == lock.rule_id
        ).with_for_update(
            nowait=nowait
        )
        rule = session.execute(stmt).scalar_one()
        logger(logging.DEBUG, 'Updating rule counters for rule %s [%d/%d/%d]' % (str(rule.id), rule.locks_ok_cnt, rule.locks_replicating_cnt, rule.locks_stuck_cnt))

        if lock.state == LockState.REPLICATING:
            rule.locks_replicating_cnt -= 1
        elif lock.state == LockState.STUCK:
            rule.locks_stuck_cnt -= 1
        rule.locks_ok_cnt += 1
        lock.state = LockState.OK
        logger(logging.DEBUG, 'Finished updating rule counters for rule %s [%d/%d/%d]' % (str(rule.id), rule.locks_ok_cnt, rule.locks_replicating_cnt, rule.locks_stuck_cnt))

        # Insert UpdatedCollectionReplica
        if rule.did_type == DIDType.DATASET:
            models.UpdatedCollectionReplica(scope=rule.scope,
                                            name=rule.name,
                                            did_type=rule.did_type,
                                            rse_id=rse_id).save(flush=False, session=session)
        elif rule.did_type == DIDType.CONTAINER:
            # Resolve to all child datasets
            for dataset in rucio.core.did.list_child_datasets(scope=rule.scope, name=rule.name, session=session):
                models.UpdatedCollectionReplica(scope=dataset['scope'],
                                                name=dataset['name'],
                                                did_type=DIDType.DATASET,
                                                rse_id=rse_id).save(flush=False, session=session)

        # Update the rule state
        if rule.state == RuleState.SUSPENDED:
            pass
        elif rule.locks_stuck_cnt > 0:
            pass
        elif rule.locks_replicating_cnt == 0 and rule.state == RuleState.REPLICATING:
            rule.state = RuleState.OK
            # Try to update the DatasetLocks
            if rule.grouping != RuleGrouping.NONE:
                stmt = select(
                    models.DatasetLock
                ).where(
                    models.DatasetLock.rule_id == rule.id
                ).with_for_update(
                    nowait=nowait
                )
                for ds_lock in session.execute(stmt).scalars().all():
                    ds_lock.state = LockState.OK
                session.flush()
            rucio.core.rule.generate_rule_notifications(rule=rule, replicating_locks_before=rule.locks_replicating_cnt + 1, session=session)
            if rule.notification == RuleNotification.YES:
                rucio.core.rule.generate_email_for_rule_ok_notification(rule=rule, session=session)
            # Try to release potential parent rules
            rucio.core.rule.release_parent_rule(child_rule_id=rule.id, session=session)
        elif rule.locks_replicating_cnt > 0 and rule.state == RuleState.REPLICATING and rule.notification == RuleNotification.PROGRESS:
            rucio.core.rule.generate_rule_notifications(rule=rule, replicating_locks_before=rule.locks_replicating_cnt + 1, session=session)

        # Insert rule history
        rucio.core.rule.insert_rule_history(rule=rule, recent=True, longterm=False, session=session)
        session.flush()


@transactional_session
def failed_transfer(scope: InternalScope, name: str, rse_id: str, error_message: Optional[str] = None, broken_rule_id: Optional[str] = None,
                    broken_message: Optional[str] = None, nowait: bool = True, *, session: "Session", logger: LoggerFunction = logging.log) -> None:
    """
    Update the state of all replica locks because of a failed transfer.
    If a transfer is permanently broken for a rule, the broken_rule_id should be filled which puts this rule into the SUSPENDED state.

    :param scope:           Scope of the did.
    :param name:            Name of the did.
    :param rse_id:          RSE id.
    :param error_message:   The error why this transfer failed.
    :param broken_rule_id:  Id of the rule which will be suspended.
    :param broken_message:  Error message for the suspended rule.
    :param nowait:          Nowait parameter for the for_update queries.
    :param session:         The database session in use.
    """

    staging_required = get_rse_attribute(rse_id, RseAttr.STAGING_REQUIRED, session=session)
    if staging_required:
        rse_name = get_rse_name(rse_id=rse_id, session=session)
        logger(logging.DEBUG, f'Destination RSE {rse_name} is type staging_required so do not update other OK replica locks.')
        stmt = select(
            models.ReplicaLock
        ).where(
            and_(models.ReplicaLock.scope == scope,
                 models.ReplicaLock.name == name,
                 models.ReplicaLock.rse_id == rse_id,
                 models.ReplicaLock.state == LockState.REPLICATING)
        ).with_for_update(
            nowait=nowait
        )
    else:
        stmt = select(
            models.ReplicaLock
        ).where(
            and_(models.ReplicaLock.scope == scope,
                 models.ReplicaLock.name == name,
                 models.ReplicaLock.rse_id == rse_id)
        ).with_for_update(
            nowait=nowait
        )

    for lock in session.execute(stmt).scalars().all():
        if lock.state == LockState.STUCK:
            continue
        logger(logging.DEBUG, 'Marking lock %s:%s for rule %s on rse %s as STUCK' % (lock.scope, lock.name, str(lock.rule_id), get_rse_name(rse_id=lock.rse_id, session=session)))
        # Update the rule counters
        stmt = select(
            models.ReplicationRule
        ).where(
            models.ReplicationRule.id == lock.rule_id
        ).with_for_update(
            nowait=nowait
        )
        rule = session.execute(stmt).scalar_one()
        logger(logging.DEBUG, 'Updating rule counters for rule %s [%d/%d/%d]' % (str(rule.id), rule.locks_ok_cnt, rule.locks_replicating_cnt, rule.locks_stuck_cnt))
        if lock.state == LockState.REPLICATING:
            rule.locks_replicating_cnt -= 1
        elif lock.state == LockState.OK:
            rule.locks_ok_cnt -= 1
        rule.locks_stuck_cnt += 1
        lock.state = LockState.STUCK
        logger(logging.DEBUG, 'Finished updating rule counters for rule %s [%d/%d/%d]' % (str(rule.id), rule.locks_ok_cnt, rule.locks_replicating_cnt, rule.locks_stuck_cnt))

        # Update the rule state
        if rule.state == RuleState.SUSPENDED:
            pass
        elif lock.rule_id == broken_rule_id:
            rule.state = RuleState.SUSPENDED
            if broken_message is not None and len(broken_message) > 245:
                rule.error = (broken_message[:245] + '...')
            else:
                rule.error = broken_message
            # Try to update the DatasetLocks
            if rule.grouping != RuleGrouping.NONE:
                stmt = select(
                    models.DatasetLock
                ).where(
                    models.DatasetLock.rule_id == rule.id
                ).with_for_update(
                    nowait=nowait
                )
                for ds_lock in session.execute(stmt).scalars().all():
                    ds_lock.state = LockState.STUCK
        elif rule.locks_stuck_cnt > 0:
            if rule.state != RuleState.STUCK:
                rule.state = RuleState.STUCK
                # Try to update the DatasetLocks
                if rule.grouping != RuleGrouping.NONE:
                    stmt = select(
                        models.DatasetLock
                    ).where(
                        models.DatasetLock.rule_id == rule.id
                    ).with_for_update(
                        nowait=nowait
                    )
                    for ds_lock in session.execute(stmt).scalars().all():
                        ds_lock.state = LockState.STUCK
            if rule.error != error_message:
                if error_message is not None and len(error_message) > 245:
                    rule.error = (error_message[:245] + '...')
                else:
                    rule.error = error_message

        # Insert rule history
        rucio.core.rule.insert_rule_history(rule=rule, recent=True, longterm=False, session=session)


@transactional_session
def touch_dataset_locks(dataset_locks: "Iterable[dict[str, Any]]", *, session: "Session") -> bool:
    """
    Update the accessed_at timestamp of the given dataset locks + eol_at.

    :param replicas: the list of dataset locks.
    :param session: The database session in use.

    :returns: True, if successful, False otherwise.
    """

    now = datetime.utcnow()
    for dataset_lock in dataset_locks:
        eol_at = define_eol(dataset_lock['scope'], dataset_lock['name'], rses=[{'id': dataset_lock['rse_id']}], session=session)
        try:
            stmt = update(
                models.DatasetLock
            ).where(
                and_(models.DatasetLock.scope == dataset_lock['scope'],
                     models.DatasetLock.name == dataset_lock['name'],
                     models.DatasetLock.rse_id == dataset_lock['rse_id'])
            ).values({
                models.DatasetLock.accessed_at: dataset_lock.get('accessed_at') or now
            }).execution_options(
                synchronize_session=False
            )
            subq = select(
                models.DatasetLock.rule_id
            ).where(
                and_(models.DatasetLock.scope == dataset_lock['scope'],
                     models.DatasetLock.name == dataset_lock['name'],
                     models.DatasetLock.rse_id == dataset_lock['rse_id'])
            )

            stmt = update(
                models.ReplicationRule
            ).where(
                models.ReplicationRule.id.in_(subq)
            ).values({
                models.ReplicationRule.eol_at: eol_at
            }).execution_options(
                synchronize_session=False
            )
            session.execute(stmt)
        except DatabaseError:
            return False

    return True
